"""
Kenshoto's Elf parser

This package will let you use programatic ninja-fu
when trying to parse Elf binaries.  The API is based
around several objects representing constructs in the
Elf binary format.  The Elf object itself contains
parsed metadata and lists of things like section headers
and relocation entries.  Additionally, most of the
objects implement repr() in some form or another which
allows you a bunch of readelf-like functionality.

*Eventually* this API will allow you to modify Elf binaries
and spit them back out in working order (not complete, you
may notice some of the initial code).

Send bug reports to rakuyo or at1as in the issue tracker
"""

# Copyright (C) 2007 Invisigoth - See LICENSE file for details
import io
import logging

from stat import *
from Elf.elf_lookup import *
import vstruct
import vstruct.defs.elf as vs_elf

logger = logging.getLogger(__name__)

HAS_STRING = [DT_NEEDED, DT_SONAME]

class Elf(vs_elf.Elf32, vs_elf.Elf64):
    def __init__(self, fd, inmem=False):
        '''
        Parse data from 'fd' and create an Elf object.

        This process attempts to get as much information from DYNAMICS as
        possible, then adds in data from SECTIONS.
        '''

        # Grab a 32bit header to use to check for other
        # machine types...
        e = vs_elf.Elf32()
        fd.seek(0)
        bytes = fd.read(len(e))
        e.vsParse(bytes)

        # if e_data == 1, then 32 bit, if e_data == 2, 64bit
        bigend = (e.e_data == ELFDATA2MSB)

        # Parse 32bit header
        if e.e_class == ELFCLASS32:
            vs_elf.Elf32.__init__(self, bigend=bigend)
            self.bits = 32
            self.psize = 4

            self._cls_reloc  = vs_elf.Elf32Reloc
            self._cls_reloca = vs_elf.Elf32Reloca
            self._cls_symbol = vs_elf.Elf32Symbol
            self._cls_section = vs_elf.Elf32Section

        # Parse 64bit header
        elif e.e_class == ELFCLASS64:
            vs_elf.Elf64.__init__(self, bigend=bigend)
            self.bits = 64
            self.psize = 8

            self._cls_reloc  = vs_elf.Elf64Reloc
            self._cls_reloca = vs_elf.Elf64Reloca
            self._cls_symbol = vs_elf.Elf64Symbol
            self._cls_section = vs_elf.Elf64Section
        else:
            raise Exception('Unrecognized e_class: %d' % e.e_class)

        self.fd = fd
        self.inmem = inmem
        self.bigend = bigend

        bytes = self.readAtOffset(0, len(self))
        self.vsParse(bytes)

        if self.e_machine == EM_386:
            self.r_types = r_types_386
        elif self.e_machine == EM_X86_64:
            self.r_types = r_types_amd64
        else:
            self.r_types = {}

        self.dyns = {}
        self.pheaders = []
        self.sections = []
        self.secnames = {}
        self.symbols  = []
        self.relocs   = []
        self.relocvas = set([])
        self.symbols_by_name = {}
        self.symbols_by_addr = {}
        self.dynamics = []      # deprecated - 2019-10-21
        self.dynamic_symbols = []
        self.dynstrtabmeta = (None, None)
        self.dynstrtab = []
        self.dynsymtabct = None     # populated by _parseDynStrs()

        try:
            logger.info('self._parsePheaders')
            self._parsePheaders()
        except Exception as e:
            logger.warning("Exception parsing Program Headers: %r" % e, exc_info=1)

        try:
            logger.info('self._parseDynLinkInfo')
            self._parseDynLinkInfo()
        except Exception as e:
            logger.warning("Exception parsing Program Headers: %r" % e, exc_info=1)


        try:
            logger.info('self._parseSections')
            self._parseSections()
        except Exception as e:
            logger.warning("Exception parsing Sections: %r" % e, exc_info=1)

        try:
            logger.info('self._parseDynamicsFromSections')
            self._parseDynamicsFromSections()
        except Exception as e:
            logger.warning("Exception parsing Dynamics from Sections: %r" % e, exc_info=1)


        # load symbols and relocs from DYNAMICS
        try:
            logger.info('self._parseDynStrs')
            self._parseDynStrs()
        except Exception as e:
            logger.warning("Exception parsing String Table (via dynamics): %r" % e, exc_info=1)

        try:
            logger.info('self._parseDynSyms')
            self._parseDynSyms()
        except Exception as e:
            logger.warning("Exception parsing Symbols (via dynamics): %r" % e, exc_info=1)

        try:
            logger.info('self._parseDynRelocs')
            self._parseDynRelocs()
        except Exception as e:
            logger.warning("Exception parsing Relocations (via dynamics): %r" % e, exc_info=1)


        # load symbols and relocs from SECTIONS
        try:
            logger.info('self._parseDynSymsFromSections')
            self._parseDynSymsFromSections()
        except Exception as e:
            logger.warning("Exception parsing Dynamic Symbols (via Sections): %r" % e, exc_info=1)

        try:
            logger.info('self._parseSectionSymbols')
            self._parseSectionSymbols()
        except Exception as e:
            logger.warning("Exception parsing Symbols (via Sections): %r" % e, exc_info=1)

        try:
            logger.info('self._parseSectionRelocs')
            self._parseSectionRelocs()
        except Exception as e:
            logger.warning("Exception parsing Relocations (via Sections): %r" % e, exc_info=1)

        logger.info('done parsing ELF')

    def __del__(self):
        try:
            self.fd.close()
        except:
            pass  # whatever. we're tearing down anyway

    def getFileBytes(self):
        '''
        Return the bytes of the file as they currently exist from the view of the file descriptor-like object

        But keeping in mind not to smash over the old location of the fd
        '''
        self.fd.flush()
        old = self.fd.tell()
        self.fd.seek(0)
        byts = self.fd.read()
        self.fd.seek(old)
        return byts

    def getRelocTypeName(self, rtype):
        '''
        Because relocation type names are decided based on the
        arch, only the Elf knows for sure...
        '''
        return self.r_types.get(rtype)

    def _parsePheaders(self):
        # Load up any program headers we find
        if self.e_phoff:
            pbase = self.e_phoff
            plen = self.e_phentsize
            for i in range(self.e_phnum):
                if self.bits == 32:
                    pgm = vs_elf.Elf32Pheader(bigend=self.bigend)
                elif self.bits == 64:
                    pgm = vs_elf.Elf64Pheader(bigend=self.bigend)
                else:
                    raise Exception('Platform not supported: %d' % (self.bits))

                bytes = self.readAtOffset(pbase, plen)
                pgm.vsParse(bytes)

                self.pheaders.append(pgm)
                pbase += plen

    def _parseSections(self):
        # Load up all the section headers
        if self.e_shoff:
            # Load up the sections
            sbase = self.e_shoff
            sec = self._cls_section(bigend=self.bigend)
            slen = self.e_shentsize
            if len(sec) != slen:
                raise Exception('Invalid Section Header Size: %d' % slen)

            secbytes = self.readAtOffset(sbase, self.e_shnum * slen)

            secs = sec * self.e_shnum
            secslen = slen * self.e_shnum
            if secslen != len(secbytes):
                logger.warning('Invalid Section-Headers Size: should be: %d   retrieved: %d', secslen, len(secbytes))

            vstruct.VArray(elems=secs).vsParse(secbytes, fast=True)

            self.sections.extend(secs)

            # Populate the section names
            strsec = self.sections[self.e_shstrndx]
            names = self.readAtOffset(strsec.sh_offset, strsec.sh_size)
            for sec in self.sections:
                name = names[sec.sh_name:].split(b"\x00")[0]
                if len(name) > 0:
                    name = name.decode('utf-8')
                    sec.setName(name)
                    self.secnames[name] = sec

    def getDynPHdr(self):
        '''
        Returns the "Dynamics" ProgramHeader
        '''
        phdr = None
        for phdr in self.getPheaders():
            if phdr.p_type == PT_DYNAMIC:
                return phdr
        return None

    def getDynBytes(self):
        '''
        Finds and returns the bytes from a Program Header of type PT_DYNAMIC
        '''
        phdr = self.getDynPHdr()
        if phdr is None:
            return None

        dynbytes = self.readAtOffset(phdr.p_offset, phdr.p_filesz)
        return dynbytes

    def _parseDynSymsFromSections(self):
        '''
        if by some strange chance, the DYNAMCS PHDR doesn't exist but we have this section...
        '''
        symtab = self.getSectionBytes('.dynsym')
        if symtab is None:
            return

        ssymtabva = self.getSection('.dynsym').sh_addr
        dsymtabva = self.dyns.get(DT_SYMTAB)
        if ssymtabva != dsymtabva:
            logger.info("Section headers and Dynamics disagree on Symbol Table: sec: 0x%x, dyn: 0x%x", ssymtabva, dsymtabva)

        # only parse the symbols that are not already accounted for.
        # symbols are ordered, so existence of index Y is always the same
        sym = self._cls_symbol(bigend=self.bigend)
        count = len(symtab) // len(sym)
        diff = count - len(self.dynamic_symbols)
        if diff == 0:
            return

        offset = len(self.dynamic_symbols) * len(sym)

        syms = sym * diff
        vstruct.VArray(elems=syms).vsParse(symtab[offset:], fast=True)

        logger.warning("_parseDynSymsFromSections:  current_count: %d\tdiff: %d\toffset: %d\t", count, diff, offset)
        for sym in syms:
            if not sym.st_name:
                continue
            name = self.getStrtabString(sym.st_name, ".dynstr")
            sym.setName(name)

            if sym in self.dynamic_symbols:
                continue
            self.dynamic_symbols.append(sym)

    def _parseDynamicsFromSections(self):
        '''
        if by some strange chance, the DYNAMCS PHDR doesn't exist but we have this section...
        '''
        dynbytes = self.getSectionBytes('.dynamic')
        while dynbytes:
            if self.bits == 32:
                dyn = vs_elf.Elf32Dynamic(bigend=self.bigend)
            elif self.bits == 64:
                dyn = vs_elf.Elf64Dynamic(bigend=self.bigend)
            else:
                raise Exception('Platform not supported: %d' % (self.bits))
            dyn.vsParse(dynbytes)

            if dyn.d_tag in HAS_STRING:
                name = self.getStrtabString(dyn.d_value, ".dynstr")
                dyn.setName(name)

            # don't add a second entry
            if dyn not in self.dynamics:
                logger.debug("dynamic: %r: 0x%x", dt_names.get(dyn.d_tag), dyn.d_value)
                self.dynamics.append(dyn)

            if dyn.d_tag == DT_NULL:  # Represents the end
                break
            dynbytes = dynbytes[len(dyn):]

    def _parseDynLinkInfo(self):
        '''
        Parse the Dynamics segment and populate both self.dynamics (legacy) and self.dyns
        This must be run before most Dynamic-data accessors like getDynStrtabString(),
        getDynSymTabInfo(), etc..
        '''
        dynbytes = self.getDynBytes()
        if dynbytes is None:
            return

        while dynbytes:
            if self.bits == 32:
                dyn = vs_elf.Elf32Dynamic(bigend=self.bigend)
            elif self.bits == 64:
                dyn = vs_elf.Elf64Dynamic(bigend=self.bigend)
            else:
                raise Exception('Platform not supported: %d' % (self.bits))

            dyn.vsParse(dynbytes)

            # dump the tag/value pairs into the "dyns" dictionary.  if multiples, create a tuple
            curdyn = self.dyns.get(dyn.d_tag)
            if curdyn is not None:
                self.dyns[dyn.d_tag] = (curdyn, dyn.d_value)
            else:
                self.dyns[dyn.d_tag] = dyn.d_value
            logger.debug('dynamic: %r: 0x%x', dt_names.get(dyn.d_tag), dyn.d_value)

            # DEPRECATED: storing info in both dyns{} and dynamics[].  
            # 2019-10-21:  dynamics will go away sometime in the future
            self.dynamics.append(dyn)
            if dyn.d_tag == DT_NULL: # Represents the end
                break
            dynbytes = dynbytes[len(dyn):]

    def _parseDynStrs(self):
        # setup STRTAB for string recovery:
        dynstrtabva = self.dyns.get(DT_STRTAB)
        strsz = self.dyns.get(DT_STRSZ)
        if dynstrtabva is None or strsz is None:
            logger.info('no dynamic string tableinfo found: DT_STRTAB: %r  DT_STRSZ: %r', dynstrtabva, strsz)
            return

        if self.dynstrtabmeta != (None, None):
            curtab = self.dynstrtabmeta[0]
            logger.warning('wtf?  multiple dynamic string tables?  old: 0x%x  new: 0x%x', curtab, rva)

        strtabbytes = self.readAtRva(dynstrtabva, strsz)

        self.dynstrtabmeta = (dynstrtabva, strsz)
        self.dynstrtab = strtabbytes.split(b'\0')

        # since our string table should certainly end in '\0', we'll have an empty string
        # at the end.  since this array is used to determine the number of symbols, we
        # need to clean it up.
        if len(self.dynstrtab) and not len(self.dynstrtab[-1]):
            self.dynstrtab.pop()

        # setup names for the dynamics table entries
        for dyn in self.dynamics:
            if dyn.d_tag in HAS_STRING:
                name = self.getDynStrtabString(dyn.d_value)
                dyn.setName(name)

    def _parseDynSyms(self):
        '''
        Parses the Symbol Table and sets up Dynamic String Table
        Using Dynamics instead of ELF Sections

        This relies on the DYNAMICS section having DT_SYMTAB and DT_SYMENT

        Because ELF has no DT_SYMTABSZ, "symtabsz" as returned from
        getDynSymTabInfo() cannot be fully trusted.  Therefore, we run a few
        sanity heuristics.

        This is only a prep run to identify symbols.  If getDynSymbol() is
        called with an index not currently in dynamic_symbols, dynamic_symbols
        is expanded to fill the need (albeit, without these sanity checks, so
        be cautious).
        '''
        # fyi:  '.dynsym' section == DT_SYMTAB
        #       '.dynstr' section == DT_STRTAB

        # parse Dynamic Symbol Table
        if len(self.dynamic_symbols):
            logger.warning('_parseDynSyms() cannot run: dynamic_symbols is not empty')
            return

        symtabrva, symsz, symtabsz = self.getDynSymTabInfo()
        if symtabrva is None:
            return

        dsoff = 0
        while True:
            syment = self.readAtRva(symtabrva + dsoff, symsz)
            sym = self._cls_symbol(bigend=self.bigend)
            sym.vsParse(syment)
            if sym.getInfoType() not in st_info_type:
                break

            if sym.getInfoBind() not in st_info_bind:
                break

            name = self.getDynStrtabString(sym.st_name)
            if name is None:
                break

            sym.setName(name)
            self.dynamic_symbols.append(sym)

            dsoff += symsz

    # FIXME: wrap in VERDEF and SYMINFO into the analysis.
    def _parseSectionSymbols(self):
        """
        Parse out the symbols that this elf binary has for us.
        """
        for sec in self.sections:
            if sec.sh_type == SHT_SYMTAB:
                sym = self._cls_symbol(bigend=self.bigend)
                symtab = self.readAtOffset(sec.sh_offset, sec.sh_size)

                count, remain = divmod(sec.sh_size, len(sym))
                syms = sym * count

                vstruct.VArray(elems=syms).vsParse(symtab, fast=True)

                for sym in syms:
                    if sym.st_name:
                        name = self.getStrtabString(sym.st_name, ".strtab")
                        sym.setName(name)
                    # logger.info('SHT_SYMTAB: %r', sym)

                    self.addSymbol(sym)

    def _parseDynRelocs(self):
        """
        Parse all the relocation entries out of Dyn table entries based at
        * REL
        * RELA
        * JMPREL
        """
        rel, relent, relsz = self.getDynRelInfo()
        if rel is not None:
            cls = self._cls_reloc
            self._doDynRelocs(rel, relsz, cls)

        rela, relaent, relasz = self.getDynRelaInfo()
        if rela is not None:
            cls = self._cls_reloca
            self._doDynRelocs(rela, relasz, cls)

        jmprel, pltrel, pltrelsz = self.getDynPltRelInfo()
        if jmprel is not None:
            cls = (self._cls_reloc, self._cls_reloca)[pltrel==DT_RELA]
            self._doDynRelocs(jmprel, pltrelsz, cls)

    def _doDynRelocs(self, rva, relsz, cls=None):
        syms = self.getDynSyms()

        if cls is None:
            cls = self._cls_reloc

        reloc = cls(bigend=self.bigend)
        relbytes = self.readAtRva(rva, relsz)
        count, remain = divmod(relsz, len(reloc))

        relocs = reloc * count
        vstruct.VArray(elems=relocs).vsParse(relbytes,fast=True)

        for reloc in relocs:
            index = reloc.getSymTabIndex()
            sym = self.getDynSymbol(index)
            if sym is not None:
                reloc.setName( sym.getName() )
            self.relocs.append((None, reloc))
            self.relocvas.add((None, reloc.r_offset))

    def _parseSectionRelocs(self):
        """
        Parse all the relocation entries out of any sections with
        sh_type == SHT_REL or SHT_RELA

        Ignores repeat relocs (ie. those already parsed from DYNAMICS)
        """
        # could it ever be interesting?  perhaps if dynamic relocs fail?
        rel, relent, relsz = self.getDynRelInfo()
        rela, relaent, relasz = self.getDynRelaInfo()
        jmprel, pltrel, pltrelsz = self.getDynPltRelInfo()
        dynrels = (rel, rela, jmprel)

        for secidx, sec in enumerate(self.sections):
            if sec.sh_type not in (SHT_REL, SHT_RELA):
                continue

            if sec.sh_offset not in dynrels:
                logger.warning('_parseSectionRelocs: Reloc section differs from Dynamics: 0x%x', sec.sh_offset)

            reloccls = self._cls_reloc
            if sec.sh_type == SHT_RELA:
                reloccls = self._cls_reloca

            secbytes = self.readAtOffset(sec.sh_offset, sec.sh_size)
            reloc = reloccls(bigend=self.bigend)
            count, remain = divmod(len(secbytes), len(reloc))

            relocs = reloc * count
            vstruct.VArray(elems=relocs).vsParse(secbytes, fast=True)

            for reloc in relocs:
                index = reloc.getSymTabIndex()
                if index < len(self.dynamic_symbols):
                    sym = self.dynamic_symbols[index]
                    reloc.setName(sym.getName())

                # key on the symbol table index because relocatable elf files use r_offset
                # as an actual offset and not a virtual address like executable images
                # For all other binaries, let dynamics win
                # Ref: https://docs.oracle.com/cd/E23824_01/html/819-0690/chapter6-54839.html 
                key = (index, reloc.r_offset)
                if key in self.relocvas or (None, reloc.r_offset) in self.relocvas:
                    logger.debug('duplicate relocation (section): %s', reloc)
                    continue

                logger.info('section reloc: %s', reloc)
                self.relocs.append((secidx, reloc))
                self.relocvas.add(key)

    def getBaseAddress(self):
        """
        For prelinked and main-exe elf binaries, return the
        value for the loaded base address...
        """
        shrd = self.isSharedObject()
        plnk = self.isPreLinked()

        # If it's a shared object and *not* prelinked,
        # we need to select a base address for it
        # FIXME find non-colliding addr in workspace
        if shrd and not plnk:
            return 0x02000000

        # Find the best base address from the list of
        # section addresses...
        base = None
        for pgm in self.getPheaders():

            if pgm.p_vaddr == 0:
                continue

            if base is None:
                base = pgm.p_vaddr
                continue

            if pgm.p_vaddr < base:
                base = pgm.p_vaddr

        if base is None:
            if self.isRelocatable():
                base = 0
            else:
                base = 0x20000000

        base &= 0xfffff000
        return base

    def readAtRva(self, rva, size):
        '''
        Calculate the file offset for the given RVA and
        read from it...
        '''
        return self.readAtOffset(self.rvaToOffset(rva), size)

    def rvaToOffset(self, rva):
        '''
        Convert an RVA for this ELF binary to a file offset.
        '''
        baseaddr = 0
        #if self.isPreLinked() or not self.isSharedObject():
        #if not self.isSharedObject():
            #logger.info('SUBTRACTING CALCULATED BASE')
            #baseaddr = self.getBaseAddress()

        for pgm in self.pheaders:
            if pgm.p_type != PT_LOAD:
                continue
            phrva = pgm.p_vaddr - baseaddr
            if rva < phrva:
                continue
            if rva >= phrva+pgm.p_memsz:
                continue
            # We are inside this pgrm header!
            rvaoff = rva - phrva
            return pgm.p_offset + rvaoff

        raise Exception(f"Couldn't convert rva: {rva} to a file offset")

    def readAtOffset(self, off, size):
        '''
        Read from the given file offset.
        '''
        self.fd.seek(off)
        return self.fd.read(size)

    def getEndian(self):
        '''
        Is architecture BigEndian?
        Returns True for MSB, False for LSB

        This works with Vivisect's definitions of ENDIAN_MSB/ENDIAN_LSB constants:
            (defined in envi/const.py)
            ENDIAN_LSB = 0
            ENDIAN_MSB = 1

        '''
        return self.e_data == ELFDATA2MSB

    def getDynRelInfo(self):
        '''
        Returns startva, size, and Entity size for any REL records
        '''
        rel = self.dyns.get(DT_REL)
        relent = self.dyns.get(DT_RELENT)
        relsz = self.dyns.get(DT_RELSZ)
        return rel, relent, relsz

    def getDynRelaInfo(self):
        '''
        Returns startva, size, and Entity size for any RELA records
        '''
        rela = self.dyns.get(DT_RELA)
        relaent = self.dyns.get(DT_RELAENT)
        relasz = self.dyns.get(DT_RELASZ)
        return rela, relaent, relasz

    def getDynPltRelInfo(self):
        '''
        Returns startva, size, and Entity size for any RELA records
        '''
        jmprel = self.dyns.get(DT_JMPREL)
        pltrel = self.dyns.get(DT_PLTREL)
        pltrelsz = self.dyns.get(DT_PLTRELSZ)
        return jmprel, pltrel, pltrelsz

    def getSection(self, secname):
        return self.secnames.get(secname, None)

    def getSectionByIndex(self, idx):
        if idx >= len(self.sections):
            return None
        return self.sections[idx]

    def getSections(self):
        """
        Return the array of sections for this Elf
        """
        return list(self.sections)

    def getSectionBytes(self, secname):
        sec = self.getSection(secname)
        if sec is None:
            return None
        return self.readAtOffset(sec.sh_offset, sec.sh_size)

    def getStrtabString(self, offset, section=".strtab"):
        sec = self.getSection(section)
        bytes = self.readAtOffset(sec.sh_offset, sec.sh_size)
        index = bytes.find(b"\x00", offset)
        return bytes[offset:index].decode('utf-8')

    def getNotes(self):
        '''
        Retrieve a list of the ElfNote vstructs from any
        sections of type SHT_NOTE.

        Example:
            for note in e.getNotes():
                print('%s : %d' % (e.name, e.ntype))
        '''
        for sec in self.getSections():
            if sec.sh_type != SHT_NOTE:
                continue

            try:
                notebytes = self.readAtOffset(sec.sh_offset, sec.sh_size)
                offset = 0
                notebyteslen = len(notebytes)
                while offset < notebyteslen:
                    note = vs_elf.ElfNote(bigend=self.getEndian())
                    if notebyteslen - offset < len(note):
                        logger.warning("""\nNOTES section length mismatch!\n\t%s
                                       \tSection Bytes: %s\n\tStranded bytes: %s\n""",
                                sec, repr(notebytes), repr(notebytes[offset:]))
                        break

                    offset = note.vsParse(notebytes, offset=offset)
                    yield note
            except Exception as e:
                logger.warning('Elf.getNotes() Exception: %s', e)

    def getPlatform(self):
        '''
        Return a "best effort" platform guess (envi platform name).
        ( and platform specific details if any )

        Example:
            plat = e.getPlatform()
        '''
        for note in self.getNotes():
            if note.name == b'GNU\x00' and note.ntype == 1:
                desc0 = int(note.desc[0])
                return osnotes.get(desc0, 'unknown')

        if self.getSection('.comment'):
            sec = self.getSection('.comment')
            if b'FreeBSD' in self.readAtOffset(sec.sh_offset, sec.sh_size):
                return 'freebsd'

        if self.getSection('QNX_info'):
            return 'qnx'

        return 'unknown'

    def getDynamics(self):
        '''
        Return a list of the dynamics.
        '''
        return list(self.dynamics)

    def getDynSyms(self):
        '''
        Return a list of dynamic symbol objects.
        '''
        return self.dynamic_symbols

    def getRelocs(self):
        '''
        Get the list of relocations.
        '''
        return list(self.relocs)

    def isPreLinked(self):
        '''
        Returns True if the Elf binary is prelinked.
        '''
        for dyn in self.dynamics:
            if dyn.d_tag == DT_GNU_PRELINKED:
                return True
            if dyn.d_tag == DT_GNU_CONFLICTSZ:
                return True
        return False

    def isSharedObject(self):
        '''
        Returns true if the given Elf binary is a dynamically shared
        object.
        '''
        if self.e_type == ET_DYN:
            return True
        return False

    def hasInterpreter(self):
        '''
        Returns true if the Elf binary has an PT_INTERP program header.
        '''
        for phdr in self.getPheaders():
            if phdr.p_type == PT_INTERP:
                return True

        return False

    def isExecutable(self):
        '''
        Returns true if the given Elf binary is an executable file type.
        Either the Elf header specifies e_type of ET_EXEC or 
        e_type == ET_DYN and has an interpreter designated in the pheaders.
        '''
        if self.e_type == ET_EXEC:
            return True

        if self.e_type == ET_DYN and self.hasInterpreter():
            return True

        return False

    def isRelocatable(self):
        '''
        Returns true if the given Elf binary is marked as a relocatable file.
        isRelocatable() helps determine if this ELF is a Kernel Module (.ko)
        or Object file (.o), *not* a Shared Object (.so) or executable.
        '''
        return self.e_type == ET_REL

    def __repr__(self, verbose=False):
        """
        Returns a string summary of this ELF.
        """
        mystr = 'Elf Binary:'
        mystr+= "\n= Intimate Details:"
        mystr+= "\n==Magic:\t\t\t\t%r"       % self.e_ident.decode('utf-8')
        mystr+= "\n==Type:\t\t\t\t\t"        + e_types.get(self.e_type)
        mystr+= "\n==Machine Arch:\t\t\t\t"  + e_machine_types.get(self.e_machine)
        mystr+= "\n==Version:\t\t\t\t%d"     % (self.e_version)
        mystr+= "\n==Entry:\t\t\t\t0x%.8x"      % (self.e_entry)
        mystr+= "\n==Program Headers(offset):\t\t%d (0x%x) bytes" % (self.e_phoff, self.e_phoff)
        mystr+= "\n==Section Headers(offset):\t\t%d (0x%x) bytes" % (self.e_shoff, self.e_shoff)
        mystr+= "\n==Flags:\t\t\t\t" + repr(self.e_flags) + " "
        mystr+= "\n==Elf Header Size:\t\t\t" + repr(self.e_ehsize) + " (" + hex(self.e_ehsize) + " bytes)"
        mystr+= "\n==Program Header Size:\t\t\t" + repr(self.e_phentsize) + " (" + hex(self.e_phentsize) + " bytes)"
        mystr+= "\n==Program Header Count:\t\t\t" + repr(self.e_phnum) + " (" + hex(self.e_phnum)+ ")"
        mystr+= "\n==Section Header Size:\t\t\t" + repr(self.e_shentsize) + " (" + hex(self.e_shentsize) + " bytes)"
        mystr+= "\n==Section Header Count:\t\t\t" + repr(self.e_shnum) + " (" + hex(self.e_shnum) + ")"
        mystr+= "\n==Section Header String Index\t\t" + repr(self.e_shstrndx) + " (" + hex(self.e_shstrndx) + " bytes)"

        mystr+= "\n\n= Sections:"
        for sec in self.sections:
            mystr+= "\n" + repr(sec)

        mystr+= "\n\n= Program Headers:"
        for ph in self.pheaders:
            mystr+= "\n" + repr(ph)

        return mystr

    def verbrepr(self):
        mystr = repr(self)

        mystr+= "\n\n= Symbols table:"
        for sym in self.symbols:
            mystr+= "\n"+repr(sym)

        mystr+= "\n\n= Relocation table:"
        for _, reloc in self.relocs:
            mystr+= "\n"+repr(reloc)

        mystr+= "\n\n= Dynamics table:"
        for dyn in self.dynamics:
            mystr+= "\n"+repr(dyn)

        mystr+= "\n\n= Dynamic Symbols table:"
        for dyn in self.dynamic_symbols:
            mystr+= "\n"+repr(dyn)

        return mystr

    def lookupSymbolName(self, name):
        """
        Lookup symbol entries in this elf binary by name.  The result is
        a long representing the address for the given symbol. Or None if
        it's not found.
        """
        return self.symbols_by_name.get(name, None)

    def lookupSymbolAddr(self, address):
        """
        lookup symbols from this elf binary by address.
        This returns the name for the given symbol or None for not found
        """
        return self.symbols_by_addr.get(address, None)

    def getPheaders(self):
        """
        Return a list of the program headers for this elf
        """
        return list(self.pheaders)

    def addSymbol(self, symbol):
        '''
        Add a symbol to the Symbols table, along with caching symbols_by_name and symbols_by_addr
        These symbols are from ELF Sections of type SHT_SYMTAB
        '''
        self.symbols.append(symbol)
        self.symbols_by_name[symbol.getName()] = symbol
        self.symbols_by_addr[symbol.st_value] = symbol

    def getSymbols(self):
        '''
        Returns discovered Symbols (from ELF Sections)
        '''
        return self.symbols

    def getSymbol(self, symidx):
        if symidx is None or symidx >= len(self.symbols):
            return None
        return self.symbols[symidx]

    def getDynSymbol(self, symidx):
        '''
        Returns the DT_SYMTAB entry at index "symidx".

        Checks self.dynamic_symbols first.
        If the index hasn't been loaded into self.dynamic_symbols, it is.
        '''
        symlen = len(self.dynamic_symbols)
        if symidx >= symlen:
            # dynamic_symbols is too small, grow
            logger.info('getDynSymbol(%d): expanding dynamic_symbols from %d', symidx, symlen)
            newspace = [self._getDynSymbol(x) for x in range(symlen, symidx + 1)]
            self.dynamic_symbols.extend(newspace)

        sym = self.dynamic_symbols[symidx]
        return sym

    def _getDynSymbol(self, symidx):
        '''
        Parse the Dynamics entries for SYMTAB and STRTAB, and return the
        symidx indexed symbol.
        '''
        symtabrva, symsz, symtabsz = self.getDynSymTabInfo()

        symrva = symtabrva + (symidx * symsz)
        # DON'T trust symtabsz.  it's often smaller than the '.dynsym' section
        # but, if we attempt to parse outside the binary, we'll throw an error.
        sym = self._cls_symbol(bigend=self.bigend)
        sym.vsParse(self.readAtRva(symrva, symsz))

        name = self.getDynStrtabString(sym.st_name)
        sym.setName(name)
        return sym

    def getDynStrTabInfo(self):
        return self.dynstrtabmeta

    def getDynSymTabInfo(self):
        '''
        Returns Symbol Table information (as obtained through Dynamics only)
        Assumes _parseDynSyms has run (populating self.dynstrtab)
        returns (symtabva, symbolsz, symtabsz)

        Because there is no DT_SYMTABSZ, we can't be certain how many dynamic
        symbols to expect.  Supposedly there is a 1:1 relationship between
        DynSyms and DynStrs, but that can be misleading.  Still, based on the
        number of DynStrs parsed in _parseDynStrs() we use that to roughly
        determine the number, which is estimated in _parseDynStrs()  and stored
        in self.dynsymtabct.  Perhaps this is horrible and should be stricken
        from the code.
        '''
        # if DT_SYMTAB doesn't exist, we don't have a symbol table defined by DYNAMICS
        symtabva = self.dyns.get(DT_SYMTAB)
        if symtabva is None:
            return None, None, None

        symsz = self.dyns.get(DT_SYMENT)

        # calculate/reuse symbol count
        if not self.dynsymtabct:    # check if we've already got this value
            # calculate the Dynamic Symbol Table Size
            self.dynsymtabct = len(self.dynstrtab)
            if not self.dynsymtabct:
                # if we haven't set up dynsymtabct, call _parseDynStrs and create one)
                self._parseDynStrs()
                self.dynsymtabct = len(self.dynstrtab)

            # if "DT_SONAME" is within this string table, there are no symbols to match that or thereafter:
            soname = self.dyns.get(DT_SONAME)
            if soname is not None and soname != -1:
                strsz = self.dyns.get(DT_STRSZ)
                if soname < strsz:
                    dynstrtabva = self.dyns.get(DT_STRTAB)
                    strtabbytes = self.readAtRva(dynstrtabva, soname)
                    dynsymstrs = strtabbytes.split(b'\0')
                    self.dynsymtabct = len(dynsymstrs) - 1

            # if any of the other Dyns (which are addresses) come after this, don't parse past them
            for dyntype in dt_rebase:
                dyn = self.dyns.get(dyntype)
                if dyn is None:
                    continue

                if dyn > symtabva:
                    if dyn < symtabva + (self.dynsymtabct * symsz):
                        # we have a new winner!
                        delta = dyn - symtabva
                        self.dynsymtabct = delta // symsz

        count = self.dynsymtabct
        symtabsz = count * symsz

        return symtabva, symsz, symtabsz

    def getDynStrtabString(self, stroff):
        '''
        Returns a string starting at stroff
        '''
        if self.dynstrtabmeta == (None, None):
            logger.info("no dyn strtabs!")
            return ''

        dynstrtabva, strsz = self.dynstrtabmeta
        strings = self.readAtRva(dynstrtabva, strsz)
        strend = strings.find(b'\0', stroff)
        if stroff > len(strings):
            return None

        return strings[stroff:strend].decode('utf-8')


def elfFromFileName(fname):
    return Elf(open(fname, 'rb'))


def elfFromBytes(fbytes):
    fd = io.BytesIO(fbytes)
    return Elf(fd)


def elfFromMemoryObject(memobj, baseaddr):
    fd = vstruct.MemObjFile(memobj, baseaddr)
    return Elf(fd)

# TODO: Consider deprecating these. We have methods on the ElfReloc objects specifically for these two
def getRelocType(val):
    return val & 0xff

def getRelocSymTabIndex(val):
    return val >> 8



STATUS_SEVERITY_SUCCESS = 0x00000000
STATUS_SEVERITY_INFORMATIONAL = 0x00000001
STATUS_SEVERITY_WARNING = 0x00000002
STATUS_SEVERITY_ERROR = 0x00000003
STATUS_WAIT_0 = 0x00000000
STATUS_WAIT_1 = 0x00000001
STATUS_WAIT_2 = 0x00000002
STATUS_WAIT_3 = 0x00000003
STATUS_WAIT_63 = 0x0000003f
STATUS_ABANDONED = 0x00000080
STATUS_ABANDONED_WAIT_0 = 0x00000080
STATUS_ABANDONED_WAIT_63 = 0x000000bf
STATUS_USER_APC = 0x000000c0
STATUS_KERNEL_APC = 0x00000100
STATUS_ALERTED = 0x00000101
STATUS_TIMEOUT = 0x00000102
STATUS_PENDING = 0x00000103
STATUS_REPARSE = 0x00000104
STATUS_MORE_ENTRIES = 0x00000105
STATUS_NOT_ALL_ASSIGNED = 0x00000106
STATUS_SOME_NOT_MAPPED = 0x00000107
STATUS_OPLOCK_BREAK_IN_PROGRESS = 0x00000108
STATUS_VOLUME_MOUNTED = 0x00000109
STATUS_RXACT_COMMITTED = 0x0000010a
STATUS_NOTIFY_CLEANUP = 0x0000010b
STATUS_NOTIFY_ENUM_DIR = 0x0000010c
STATUS_NO_QUOTAS_FOR_ACCOUNT = 0x0000010d
STATUS_PRIMARY_TRANSPORT_CONNECT_FAILED = 0x0000010e
STATUS_PAGE_FAULT_TRANSITION = 0x00000110
STATUS_PAGE_FAULT_DEMAND_ZERO = 0x00000111
STATUS_PAGE_FAULT_COPY_ON_WRITE = 0x00000112
STATUS_PAGE_FAULT_GUARD_PAGE = 0x00000113
STATUS_PAGE_FAULT_PAGING_FILE = 0x00000114
STATUS_CACHE_PAGE_LOCKED = 0x00000115
STATUS_CRASH_DUMP = 0x00000116
STATUS_BUFFER_ALL_ZEROS = 0x00000117
STATUS_REPARSE_OBJECT = 0x00000118
STATUS_RESOURCE_REQUIREMENTS_CHANGED = 0x00000119
STATUS_TRANSLATION_COMPLETE = 0x00000120
STATUS_DS_MEMBERSHIP_EVALUATED_LOCALLY = 0x00000121
STATUS_NOTHING_TO_TERMINATE = 0x00000122
STATUS_PROCESS_NOT_IN_JOB = 0x00000123
STATUS_PROCESS_IN_JOB = 0x00000124
STATUS_VOLSNAP_HIBERNATE_READY = 0x00000125
STATUS_FSFILTER_OP_COMPLETED_SUCCESSFULLY = 0x00000126
STATUS_OBJECT_NAME_EXISTS = 0x40000000
STATUS_THREAD_WAS_SUSPENDED = 0x40000001
STATUS_WORKING_SET_LIMIT_RANGE = 0x40000002
STATUS_IMAGE_NOT_AT_BASE = 0x40000003
STATUS_RXACT_STATE_CREATED = 0x40000004
STATUS_SEGMENT_NOTIFICATION = 0x40000005
STATUS_LOCAL_USER_SESSION_KEY = 0x40000006
STATUS_BAD_CURRENT_DIRECTORY = 0x40000007
STATUS_SERIAL_MORE_WRITES = 0x40000008
STATUS_REGISTRY_RECOVERED = 0x40000009
STATUS_FT_READ_RECOVERY_FROM_BACKUP = 0x4000000a
STATUS_FT_WRITE_RECOVERY = 0x4000000b
STATUS_SERIAL_COUNTER_TIMEOUT = 0x4000000c
STATUS_NULL_LM_PASSWORD = 0x4000000d
STATUS_IMAGE_MACHINE_TYPE_MISMATCH = 0x4000000e
STATUS_RECEIVE_PARTIAL = 0x4000000f
STATUS_RECEIVE_EXPEDITED = 0x40000010
STATUS_RECEIVE_PARTIAL_EXPEDITED = 0x40000011
STATUS_EVENT_DONE = 0x40000012
STATUS_EVENT_PENDING = 0x40000013
STATUS_CHECKING_FILE_SYSTEM = 0x40000014
STATUS_FATAL_APP_EXIT = 0x40000015
STATUS_PREDEFINED_HANDLE = 0x40000016
STATUS_WAS_UNLOCKED = 0x40000017
STATUS_SERVICE_NOTIFICATION = 0x40000018
STATUS_WAS_LOCKED = 0x40000019
STATUS_LOG_HARD_ERROR = 0x4000001a
STATUS_ALREADY_WIN32 = 0x4000001b
STATUS_WX86_UNSIMULATE = 0x4000001c
STATUS_WX86_CONTINUE = 0x4000001d
STATUS_WX86_SINGLE_STEP = 0x4000001e
STATUS_WX86_BREAKPOINT = 0x4000001f
STATUS_WX86_EXCEPTION_CONTINUE = 0x40000020
STATUS_WX86_EXCEPTION_LASTCHANCE = 0x40000021
STATUS_WX86_EXCEPTION_CHAIN = 0x40000022
STATUS_IMAGE_MACHINE_TYPE_MISMATCH_EXE = 0x40000023
STATUS_NO_YIELD_PERFORMED = 0x40000024
STATUS_TIMER_RESUME_IGNORED = 0x40000025
STATUS_ARBITRATION_UNHANDLED = 0x40000026
STATUS_CARDBUS_NOT_SUPPORTED = 0x40000027
STATUS_WX86_CREATEWX86TIB = 0x40000028
STATUS_MP_PROCESSOR_MISMATCH = 0x40000029
STATUS_HIBERNATED = 0x4000002a
STATUS_RESUME_HIBERNATION = 0x4000002b
STATUS_FIRMWARE_UPDATED = 0x4000002c
STATUS_WAKE_SYSTEM = 0x40000294
STATUS_DS_SHUTTING_DOWN = 0x40000370
RPC_NT_UUID_LOCAL_ONLY = 0x40020056
RPC_NT_SEND_INCOMPLETE = 0x400200af
STATUS_CTX_CDM_CONNECT = 0x400a0004
STATUS_CTX_CDM_DISCONNECT = 0x400a0005
STATUS_SXS_RELEASE_ACTIVATION_CONTEXT = 0x4015000d
STATUS_GUARD_PAGE_VIOLATION = 0x80000001
STATUS_DATATYPE_MISALIGNMENT = 0x80000002
STATUS_BREAKPOINT = 0x80000003
STATUS_SINGLE_STEP = 0x80000004
STATUS_BUFFER_OVERFLOW = 0x80000005
STATUS_NO_MORE_FILES = 0x80000006
STATUS_WAKE_SYSTEM_DEBUGGER = 0x80000007
STATUS_HANDLES_CLOSED = 0x8000000a
STATUS_NO_INHERITANCE = 0x8000000b
STATUS_GUID_SUBSTITUTION_MADE = 0x8000000c
STATUS_PARTIAL_COPY = 0x8000000d
STATUS_DEVICE_PAPER_EMPTY = 0x8000000e
STATUS_DEVICE_POWERED_OFF = 0x8000000f
STATUS_DEVICE_OFF_LINE = 0x80000010
STATUS_DEVICE_BUSY = 0x80000011
STATUS_NO_MORE_EAS = 0x80000012
STATUS_INVALID_EA_NAME = 0x80000013
STATUS_EA_LIST_INCONSISTENT = 0x80000014
STATUS_INVALID_EA_FLAG = 0x80000015
STATUS_VERIFY_REQUIRED = 0x80000016
STATUS_EXTRANEOUS_INFORMATION = 0x80000017
STATUS_RXACT_COMMIT_NECESSARY = 0x80000018
STATUS_NO_MORE_ENTRIES = 0x8000001a
STATUS_FILEMARK_DETECTED = 0x8000001b
STATUS_MEDIA_CHANGED = 0x8000001c
STATUS_BUS_RESET = 0x8000001d
STATUS_END_OF_MEDIA = 0x8000001e
STATUS_BEGINNING_OF_MEDIA = 0x8000001f
STATUS_MEDIA_CHECK = 0x80000020
STATUS_SETMARK_DETECTED = 0x80000021
STATUS_NO_DATA_DETECTED = 0x80000022
STATUS_REDIRECTOR_HAS_OPEN_HANDLES = 0x80000023
STATUS_SERVER_HAS_OPEN_HANDLES = 0x80000024
STATUS_ALREADY_DISCONNECTED = 0x80000025
STATUS_LONGJUMP = 0x80000026
STATUS_CLEANER_CARTRIDGE_INSTALLED = 0x80000027
STATUS_PLUGPLAY_QUERY_VETOED = 0x80000028
STATUS_UNWIND_CONSOLIDATE = 0x80000029
STATUS_REGISTRY_HIVE_RECOVERED = 0x8000002a
STATUS_DLL_MIGHT_BE_INSECURE = 0x8000002b
STATUS_DLL_MIGHT_BE_INCOMPATIBLE = 0x8000002c
STATUS_DEVICE_REQUIRES_CLEANING = 0x80000288
STATUS_DEVICE_DOOR_OPEN = 0x80000289
STATUS_CLUSTER_NODE_ALREADY_UP = 0x80130001
STATUS_CLUSTER_NODE_ALREADY_DOWN = 0x80130002
STATUS_CLUSTER_NETWORK_ALREADY_ONLINE = 0x80130003
STATUS_CLUSTER_NETWORK_ALREADY_OFFLINE = 0x80130004
STATUS_CLUSTER_NODE_ALREADY_MEMBER = 0x80130005
STATUS_UNSUCCESSFUL = 0xc0000001
STATUS_NOT_IMPLEMENTED = 0xc0000002
STATUS_INVALID_INFO_CLASS = 0xc0000003
STATUS_INFO_LENGTH_MISMATCH = 0xc0000004
STATUS_ACCESS_VIOLATION = 0xc0000005
STATUS_IN_PAGE_ERROR = 0xc0000006
STATUS_PAGEFILE_QUOTA = 0xc0000007
STATUS_INVALID_HANDLE = 0xc0000008
STATUS_BAD_INITIAL_STACK = 0xc0000009
STATUS_BAD_INITIAL_PC = 0xc000000a
STATUS_INVALID_CID = 0xc000000b
STATUS_TIMER_NOT_CANCELED = 0xc000000c
STATUS_INVALID_PARAMETER = 0xc000000d
STATUS_NO_SUCH_DEVICE = 0xc000000e
STATUS_NO_SUCH_FILE = 0xc000000f
STATUS_INVALID_DEVICE_REQUEST = 0xc0000010
STATUS_END_OF_FILE = 0xc0000011
STATUS_WRONG_VOLUME = 0xc0000012
STATUS_NO_MEDIA_IN_DEVICE = 0xc0000013
STATUS_UNRECOGNIZED_MEDIA = 0xc0000014
STATUS_NONEXISTENT_SECTOR = 0xc0000015
STATUS_MORE_PROCESSING_REQUIRED = 0xc0000016
STATUS_NO_MEMORY = 0xc0000017
STATUS_CONFLICTING_ADDRESSES = 0xc0000018
STATUS_NOT_MAPPED_VIEW = 0xc0000019
STATUS_UNABLE_TO_FREE_VM = 0xc000001a
STATUS_UNABLE_TO_DELETE_SECTION = 0xc000001b
STATUS_INVALID_SYSTEM_SERVICE = 0xc000001c
STATUS_ILLEGAL_INSTRUCTION = 0xc000001d
STATUS_INVALID_LOCK_SEQUENCE = 0xc000001e
STATUS_INVALID_VIEW_SIZE = 0xc000001f
STATUS_INVALID_FILE_FOR_SECTION = 0xc0000020
STATUS_ALREADY_COMMITTED = 0xc0000021
STATUS_ACCESS_DENIED = 0xc0000022
STATUS_BUFFER_TOO_SMALL = 0xc0000023
STATUS_OBJECT_TYPE_MISMATCH = 0xc0000024
STATUS_NONCONTINUABLE_EXCEPTION = 0xc0000025
STATUS_INVALID_DISPOSITION = 0xc0000026
STATUS_UNWIND = 0xc0000027
STATUS_BAD_STACK = 0xc0000028
STATUS_INVALID_UNWIND_TARGET = 0xc0000029
STATUS_NOT_LOCKED = 0xc000002a
STATUS_PARITY_ERROR = 0xc000002b
STATUS_UNABLE_TO_DECOMMIT_VM = 0xc000002c
STATUS_NOT_COMMITTED = 0xc000002d
STATUS_INVALID_PORT_ATTRIBUTES = 0xc000002e
STATUS_PORT_MESSAGE_TOO_LONG = 0xc000002f
STATUS_INVALID_PARAMETER_MIX = 0xc0000030
STATUS_INVALID_QUOTA_LOWER = 0xc0000031
STATUS_DISK_CORRUPT_ERROR = 0xc0000032
STATUS_OBJECT_NAME_INVALID = 0xc0000033
STATUS_OBJECT_NAME_NOT_FOUND = 0xc0000034
STATUS_OBJECT_NAME_COLLISION = 0xc0000035
STATUS_PORT_DISCONNECTED = 0xc0000037
STATUS_DEVICE_ALREADY_ATTACHED = 0xc0000038
STATUS_OBJECT_PATH_INVALID = 0xc0000039
STATUS_OBJECT_PATH_NOT_FOUND = 0xc000003a
STATUS_OBJECT_PATH_SYNTAX_BAD = 0xc000003b
STATUS_DATA_OVERRUN = 0xc000003c
STATUS_DATA_LATE_ERROR = 0xc000003d
STATUS_DATA_ERROR = 0xc000003e
STATUS_CRC_ERROR = 0xc000003f
STATUS_SECTION_TOO_BIG = 0xc0000040
STATUS_PORT_CONNECTION_REFUSED = 0xc0000041
STATUS_INVALID_PORT_HANDLE = 0xc0000042
STATUS_SHARING_VIOLATION = 0xc0000043
STATUS_QUOTA_EXCEEDED = 0xc0000044
STATUS_INVALID_PAGE_PROTECTION = 0xc0000045
STATUS_MUTANT_NOT_OWNED = 0xc0000046
STATUS_SEMAPHORE_LIMIT_EXCEEDED = 0xc0000047
STATUS_PORT_ALREADY_SET = 0xc0000048
STATUS_SECTION_NOT_IMAGE = 0xc0000049
STATUS_SUSPEND_COUNT_EXCEEDED = 0xc000004a
STATUS_THREAD_IS_TERMINATING = 0xc000004b
STATUS_BAD_WORKING_SET_LIMIT = 0xc000004c
STATUS_INCOMPATIBLE_FILE_MAP = 0xc000004d
STATUS_SECTION_PROTECTION = 0xc000004e
STATUS_EAS_NOT_SUPPORTED = 0xc000004f
STATUS_EA_TOO_LARGE = 0xc0000050
STATUS_NONEXISTENT_EA_ENTRY = 0xc0000051
STATUS_NO_EAS_ON_FILE = 0xc0000052
STATUS_EA_CORRUPT_ERROR = 0xc0000053
STATUS_FILE_LOCK_CONFLICT = 0xc0000054
STATUS_LOCK_NOT_GRANTED = 0xc0000055
STATUS_DELETE_PENDING = 0xc0000056
STATUS_CTL_FILE_NOT_SUPPORTED = 0xc0000057
STATUS_UNKNOWN_REVISION = 0xc0000058
STATUS_REVISION_MISMATCH = 0xc0000059
STATUS_INVALID_OWNER = 0xc000005a
STATUS_INVALID_PRIMARY_GROUP = 0xc000005b
STATUS_NO_IMPERSONATION_TOKEN = 0xc000005c
STATUS_CANT_DISABLE_MANDATORY = 0xc000005d
STATUS_NO_LOGON_SERVERS = 0xc000005e
STATUS_NO_SUCH_LOGON_SESSION = 0xc000005f
STATUS_NO_SUCH_PRIVILEGE = 0xc0000060
STATUS_PRIVILEGE_NOT_HELD = 0xc0000061
STATUS_INVALID_ACCOUNT_NAME = 0xc0000062
STATUS_USER_EXISTS = 0xc0000063
STATUS_NO_SUCH_USER = 0xc0000064
STATUS_GROUP_EXISTS = 0xc0000065
STATUS_NO_SUCH_GROUP = 0xc0000066
STATUS_MEMBER_IN_GROUP = 0xc0000067
STATUS_MEMBER_NOT_IN_GROUP = 0xc0000068
STATUS_LAST_ADMIN = 0xc0000069
STATUS_WRONG_PASSWORD = 0xc000006a
STATUS_ILL_FORMED_PASSWORD = 0xc000006b
STATUS_PASSWORD_RESTRICTION = 0xc000006c
STATUS_LOGON_FAILURE = 0xc000006d
STATUS_ACCOUNT_RESTRICTION = 0xc000006e
STATUS_INVALID_LOGON_HOURS = 0xc000006f
STATUS_INVALID_WORKSTATION = 0xc0000070
STATUS_PASSWORD_EXPIRED = 0xc0000071
STATUS_ACCOUNT_DISABLED = 0xc0000072
STATUS_NONE_MAPPED = 0xc0000073
STATUS_TOO_MANY_LUIDS_REQUESTED = 0xc0000074
STATUS_LUIDS_EXHAUSTED = 0xc0000075
STATUS_INVALID_SUB_AUTHORITY = 0xc0000076
STATUS_INVALID_ACL = 0xc0000077
STATUS_INVALID_SID = 0xc0000078
STATUS_INVALID_SECURITY_DESCR = 0xc0000079
STATUS_PROCEDURE_NOT_FOUND = 0xc000007a
STATUS_INVALID_IMAGE_FORMAT = 0xc000007b
STATUS_NO_TOKEN = 0xc000007c
STATUS_BAD_INHERITANCE_ACL = 0xc000007d
STATUS_RANGE_NOT_LOCKED = 0xc000007e
STATUS_DISK_FULL = 0xc000007f
STATUS_SERVER_DISABLED = 0xc0000080
STATUS_SERVER_NOT_DISABLED = 0xc0000081
STATUS_TOO_MANY_GUIDS_REQUESTED = 0xc0000082
STATUS_GUIDS_EXHAUSTED = 0xc0000083
STATUS_INVALID_ID_AUTHORITY = 0xc0000084
STATUS_AGENTS_EXHAUSTED = 0xc0000085
STATUS_INVALID_VOLUME_LABEL = 0xc0000086
STATUS_SECTION_NOT_EXTENDED = 0xc0000087
STATUS_NOT_MAPPED_DATA = 0xc0000088
STATUS_RESOURCE_DATA_NOT_FOUND = 0xc0000089
STATUS_RESOURCE_TYPE_NOT_FOUND = 0xc000008a
STATUS_RESOURCE_NAME_NOT_FOUND = 0xc000008b
STATUS_ARRAY_BOUNDS_EXCEEDED = 0xc000008c
STATUS_FLOAT_DENORMAL_OPERAND = 0xc000008d
STATUS_FLOAT_DIVIDE_BY_ZERO = 0xc000008e
STATUS_FLOAT_INEXACT_RESULT = 0xc000008f
STATUS_FLOAT_INVALID_OPERATION = 0xc0000090
STATUS_FLOAT_OVERFLOW = 0xc0000091
STATUS_FLOAT_STACK_CHECK = 0xc0000092
STATUS_FLOAT_UNDERFLOW = 0xc0000093
STATUS_INTEGER_DIVIDE_BY_ZERO = 0xc0000094
STATUS_INTEGER_OVERFLOW = 0xc0000095
STATUS_PRIVILEGED_INSTRUCTION = 0xc0000096
STATUS_TOO_MANY_PAGING_FILES = 0xc0000097
STATUS_FILE_INVALID = 0xc0000098
STATUS_ALLOTTED_SPACE_EXCEEDED = 0xc0000099
STATUS_INSUFFICIENT_RESOURCES = 0xc000009a
STATUS_DFS_EXIT_PATH_FOUND = 0xc000009b
STATUS_DEVICE_DATA_ERROR = 0xc000009c
STATUS_DEVICE_NOT_CONNECTED = 0xc000009d
STATUS_DEVICE_POWER_FAILURE = 0xc000009e
STATUS_FREE_VM_NOT_AT_BASE = 0xc000009f
STATUS_MEMORY_NOT_ALLOCATED = 0xc00000a0
STATUS_WORKING_SET_QUOTA = 0xc00000a1
STATUS_MEDIA_WRITE_PROTECTED = 0xc00000a2
STATUS_DEVICE_NOT_READY = 0xc00000a3
STATUS_INVALID_GROUP_ATTRIBUTES = 0xc00000a4
STATUS_BAD_IMPERSONATION_LEVEL = 0xc00000a5
STATUS_CANT_OPEN_ANONYMOUS = 0xc00000a6
STATUS_BAD_VALIDATION_CLASS = 0xc00000a7
STATUS_BAD_TOKEN_TYPE = 0xc00000a8
STATUS_BAD_MASTER_BOOT_RECORD = 0xc00000a9
STATUS_INSTRUCTION_MISALIGNMENT = 0xc00000aa
STATUS_INSTANCE_NOT_AVAILABLE = 0xc00000ab
STATUS_PIPE_NOT_AVAILABLE = 0xc00000ac
STATUS_INVALID_PIPE_STATE = 0xc00000ad
STATUS_PIPE_BUSY = 0xc00000ae
STATUS_ILLEGAL_FUNCTION = 0xc00000af
STATUS_PIPE_DISCONNECTED = 0xc00000b0
STATUS_PIPE_CLOSING = 0xc00000b1
STATUS_PIPE_CONNECTED = 0xc00000b2
STATUS_PIPE_LISTENING = 0xc00000b3
STATUS_INVALID_READ_MODE = 0xc00000b4
STATUS_IO_TIMEOUT = 0xc00000b5
STATUS_FILE_FORCED_CLOSED = 0xc00000b6
STATUS_PROFILING_NOT_STARTED = 0xc00000b7
STATUS_PROFILING_NOT_STOPPED = 0xc00000b8
STATUS_COULD_NOT_INTERPRET = 0xc00000b9
STATUS_FILE_IS_A_DIRECTORY = 0xc00000ba
STATUS_NOT_SUPPORTED = 0xc00000bb
STATUS_REMOTE_NOT_LISTENING = 0xc00000bc
STATUS_DUPLICATE_NAME = 0xc00000bd
STATUS_BAD_NETWORK_PATH = 0xc00000be
STATUS_NETWORK_BUSY = 0xc00000bf
STATUS_DEVICE_DOES_NOT_EXIST = 0xc00000c0
STATUS_TOO_MANY_COMMANDS = 0xc00000c1
STATUS_ADAPTER_HARDWARE_ERROR = 0xc00000c2
STATUS_INVALID_NETWORK_RESPONSE = 0xc00000c3
STATUS_UNEXPECTED_NETWORK_ERROR = 0xc00000c4
STATUS_BAD_REMOTE_ADAPTER = 0xc00000c5
STATUS_PRINT_QUEUE_FULL = 0xc00000c6
STATUS_NO_SPOOL_SPACE = 0xc00000c7
STATUS_PRINT_CANCELLED = 0xc00000c8
STATUS_NETWORK_NAME_DELETED = 0xc00000c9
STATUS_NETWORK_ACCESS_DENIED = 0xc00000ca
STATUS_BAD_DEVICE_TYPE = 0xc00000cb
STATUS_BAD_NETWORK_NAME = 0xc00000cc
STATUS_TOO_MANY_NAMES = 0xc00000cd
STATUS_TOO_MANY_SESSIONS = 0xc00000ce
STATUS_SHARING_PAUSED = 0xc00000cf
STATUS_REQUEST_NOT_ACCEPTED = 0xc00000d0
STATUS_REDIRECTOR_PAUSED = 0xc00000d1
STATUS_NET_WRITE_FAULT = 0xc00000d2
STATUS_PROFILING_AT_LIMIT = 0xc00000d3
STATUS_NOT_SAME_DEVICE = 0xc00000d4
STATUS_FILE_RENAMED = 0xc00000d5
STATUS_VIRTUAL_CIRCUIT_CLOSED = 0xc00000d6
STATUS_NO_SECURITY_ON_OBJECT = 0xc00000d7
STATUS_CANT_WAIT = 0xc00000d8
STATUS_PIPE_EMPTY = 0xc00000d9
STATUS_CANT_ACCESS_DOMAIN_INFO = 0xc00000da
STATUS_CANT_TERMINATE_SELF = 0xc00000db
STATUS_INVALID_SERVER_STATE = 0xc00000dc
STATUS_INVALID_DOMAIN_STATE = 0xc00000dd
STATUS_INVALID_DOMAIN_ROLE = 0xc00000de
STATUS_NO_SUCH_DOMAIN = 0xc00000df
STATUS_DOMAIN_EXISTS = 0xc00000e0
STATUS_DOMAIN_LIMIT_EXCEEDED = 0xc00000e1
STATUS_OPLOCK_NOT_GRANTED = 0xc00000e2
STATUS_INVALID_OPLOCK_PROTOCOL = 0xc00000e3
STATUS_INTERNAL_DB_CORRUPTION = 0xc00000e4
STATUS_INTERNAL_ERROR = 0xc00000e5
STATUS_GENERIC_NOT_MAPPED = 0xc00000e6
STATUS_BAD_DESCRIPTOR_FORMAT = 0xc00000e7
STATUS_INVALID_USER_BUFFER = 0xc00000e8
STATUS_UNEXPECTED_IO_ERROR = 0xc00000e9
STATUS_UNEXPECTED_MM_CREATE_ERR = 0xc00000ea
STATUS_UNEXPECTED_MM_MAP_ERROR = 0xc00000eb
STATUS_UNEXPECTED_MM_EXTEND_ERR = 0xc00000ec
STATUS_NOT_LOGON_PROCESS = 0xc00000ed
STATUS_LOGON_SESSION_EXISTS = 0xc00000ee
STATUS_INVALID_PARAMETER_1 = 0xc00000ef
STATUS_INVALID_PARAMETER_2 = 0xc00000f0
STATUS_INVALID_PARAMETER_3 = 0xc00000f1
STATUS_INVALID_PARAMETER_4 = 0xc00000f2
STATUS_INVALID_PARAMETER_5 = 0xc00000f3
STATUS_INVALID_PARAMETER_6 = 0xc00000f4
STATUS_INVALID_PARAMETER_7 = 0xc00000f5
STATUS_INVALID_PARAMETER_8 = 0xc00000f6
STATUS_INVALID_PARAMETER_9 = 0xc00000f7
STATUS_INVALID_PARAMETER_10 = 0xc00000f8
STATUS_INVALID_PARAMETER_11 = 0xc00000f9
STATUS_INVALID_PARAMETER_12 = 0xc00000fa
STATUS_REDIRECTOR_NOT_STARTED = 0xc00000fb
STATUS_REDIRECTOR_STARTED = 0xc00000fc
STATUS_STACK_OVERFLOW = 0xc00000fd
STATUS_NO_SUCH_PACKAGE = 0xc00000fe
STATUS_BAD_FUNCTION_TABLE = 0xc00000ff
STATUS_VARIABLE_NOT_FOUND = 0xc0000100
STATUS_DIRECTORY_NOT_EMPTY = 0xc0000101
STATUS_FILE_CORRUPT_ERROR = 0xc0000102
STATUS_NOT_A_DIRECTORY = 0xc0000103
STATUS_BAD_LOGON_SESSION_STATE = 0xc0000104
STATUS_LOGON_SESSION_COLLISION = 0xc0000105
STATUS_NAME_TOO_LONG = 0xc0000106
STATUS_FILES_OPEN = 0xc0000107
STATUS_CONNECTION_IN_USE = 0xc0000108
STATUS_MESSAGE_NOT_FOUND = 0xc0000109
STATUS_PROCESS_IS_TERMINATING = 0xc000010a
STATUS_INVALID_LOGON_TYPE = 0xc000010b
STATUS_NO_GUID_TRANSLATION = 0xc000010c
STATUS_CANNOT_IMPERSONATE = 0xc000010d
STATUS_IMAGE_ALREADY_LOADED = 0xc000010e
STATUS_ABIOS_NOT_PRESENT = 0xc000010f
STATUS_ABIOS_LID_NOT_EXIST = 0xc0000110
STATUS_ABIOS_LID_ALREADY_OWNED = 0xc0000111
STATUS_ABIOS_NOT_LID_OWNER = 0xc0000112
STATUS_ABIOS_INVALID_COMMAND = 0xc0000113
STATUS_ABIOS_INVALID_LID = 0xc0000114
STATUS_ABIOS_SELECTOR_NOT_AVAILABLE = 0xc0000115
STATUS_ABIOS_INVALID_SELECTOR = 0xc0000116
STATUS_NO_LDT = 0xc0000117
STATUS_INVALID_LDT_SIZE = 0xc0000118
STATUS_INVALID_LDT_OFFSET = 0xc0000119
STATUS_INVALID_LDT_DESCRIPTOR = 0xc000011a
STATUS_INVALID_IMAGE_NE_FORMAT = 0xc000011b
STATUS_RXACT_INVALID_STATE = 0xc000011c
STATUS_RXACT_COMMIT_FAILURE = 0xc000011d
STATUS_MAPPED_FILE_SIZE_ZERO = 0xc000011e
STATUS_TOO_MANY_OPENED_FILES = 0xc000011f
STATUS_CANCELLED = 0xc0000120
STATUS_CANNOT_DELETE = 0xc0000121
STATUS_INVALID_COMPUTER_NAME = 0xc0000122
STATUS_FILE_DELETED = 0xc0000123
STATUS_SPECIAL_ACCOUNT = 0xc0000124
STATUS_SPECIAL_GROUP = 0xc0000125
STATUS_SPECIAL_USER = 0xc0000126
STATUS_MEMBERS_PRIMARY_GROUP = 0xc0000127
STATUS_FILE_CLOSED = 0xc0000128
STATUS_TOO_MANY_THREADS = 0xc0000129
STATUS_THREAD_NOT_IN_PROCESS = 0xc000012a
STATUS_TOKEN_ALREADY_IN_USE = 0xc000012b
STATUS_PAGEFILE_QUOTA_EXCEEDED = 0xc000012c
STATUS_COMMITMENT_LIMIT = 0xc000012d
STATUS_INVALID_IMAGE_LE_FORMAT = 0xc000012e
STATUS_INVALID_IMAGE_NOT_MZ = 0xc000012f
STATUS_INVALID_IMAGE_PROTECT = 0xc0000130
STATUS_INVALID_IMAGE_WIN_16 = 0xc0000131
STATUS_LOGON_SERVER_CONFLICT = 0xc0000132
STATUS_TIME_DIFFERENCE_AT_DC = 0xc0000133
STATUS_SYNCHRONIZATION_REQUIRED = 0xc0000134
STATUS_DLL_NOT_FOUND = 0xc0000135
STATUS_OPEN_FAILED = 0xc0000136
STATUS_IO_PRIVILEGE_FAILED = 0xc0000137
STATUS_ORDINAL_NOT_FOUND = 0xc0000138
STATUS_ENTRYPOINT_NOT_FOUND = 0xc0000139
STATUS_CONTROL_C_EXIT = 0xc000013a
STATUS_LOCAL_DISCONNECT = 0xc000013b
STATUS_REMOTE_DISCONNECT = 0xc000013c
STATUS_REMOTE_RESOURCES = 0xc000013d
STATUS_LINK_FAILED = 0xc000013e
STATUS_LINK_TIMEOUT = 0xc000013f
STATUS_INVALID_CONNECTION = 0xc0000140
STATUS_INVALID_ADDRESS = 0xc0000141
STATUS_DLL_INIT_FAILED = 0xc0000142
STATUS_MISSING_SYSTEMFILE = 0xc0000143
STATUS_UNHANDLED_EXCEPTION = 0xc0000144
STATUS_APP_INIT_FAILURE = 0xc0000145
STATUS_PAGEFILE_CREATE_FAILED = 0xc0000146
STATUS_NO_PAGEFILE = 0xc0000147
STATUS_INVALID_LEVEL = 0xc0000148
STATUS_WRONG_PASSWORD_CORE = 0xc0000149
STATUS_ILLEGAL_FLOAT_CONTEXT = 0xc000014a
STATUS_PIPE_BROKEN = 0xc000014b
STATUS_REGISTRY_CORRUPT = 0xc000014c
STATUS_REGISTRY_IO_FAILED = 0xc000014d
STATUS_NO_EVENT_PAIR = 0xc000014e
STATUS_UNRECOGNIZED_VOLUME = 0xc000014f
STATUS_SERIAL_NO_DEVICE_INITED = 0xc0000150
STATUS_NO_SUCH_ALIAS = 0xc0000151
STATUS_MEMBER_NOT_IN_ALIAS = 0xc0000152
STATUS_MEMBER_IN_ALIAS = 0xc0000153
STATUS_ALIAS_EXISTS = 0xc0000154
STATUS_LOGON_NOT_GRANTED = 0xc0000155
STATUS_TOO_MANY_SECRETS = 0xc0000156
STATUS_SECRET_TOO_LONG = 0xc0000157
STATUS_INTERNAL_DB_ERROR = 0xc0000158
STATUS_FULLSCREEN_MODE = 0xc0000159
STATUS_TOO_MANY_CONTEXT_IDS = 0xc000015a
STATUS_LOGON_TYPE_NOT_GRANTED = 0xc000015b
STATUS_NOT_REGISTRY_FILE = 0xc000015c
STATUS_NT_CROSS_ENCRYPTION_REQUIRED = 0xc000015d
STATUS_DOMAIN_CTRLR_CONFIG_ERROR = 0xc000015e
STATUS_FT_MISSING_MEMBER = 0xc000015f
STATUS_ILL_FORMED_SERVICE_ENTRY = 0xc0000160
STATUS_ILLEGAL_CHARACTER = 0xc0000161
STATUS_UNMAPPABLE_CHARACTER = 0xc0000162
STATUS_UNDEFINED_CHARACTER = 0xc0000163
STATUS_FLOPPY_VOLUME = 0xc0000164
STATUS_FLOPPY_ID_MARK_NOT_FOUND = 0xc0000165
STATUS_FLOPPY_WRONG_CYLINDER = 0xc0000166
STATUS_FLOPPY_UNKNOWN_ERROR = 0xc0000167
STATUS_FLOPPY_BAD_REGISTERS = 0xc0000168
STATUS_DISK_RECALIBRATE_FAILED = 0xc0000169
STATUS_DISK_OPERATION_FAILED = 0xc000016a
STATUS_DISK_RESET_FAILED = 0xc000016b
STATUS_SHARED_IRQ_BUSY = 0xc000016c
STATUS_FT_ORPHANING = 0xc000016d
STATUS_BIOS_FAILED_TO_CONNECT_INTERRUPT = 0xc000016e
STATUS_PARTITION_FAILURE = 0xc0000172
STATUS_INVALID_BLOCK_LENGTH = 0xc0000173
STATUS_DEVICE_NOT_PARTITIONED = 0xc0000174
STATUS_UNABLE_TO_LOCK_MEDIA = 0xc0000175
STATUS_UNABLE_TO_UNLOAD_MEDIA = 0xc0000176
STATUS_EOM_OVERFLOW = 0xc0000177
STATUS_NO_MEDIA = 0xc0000178
STATUS_NO_SUCH_MEMBER = 0xc000017a
STATUS_INVALID_MEMBER = 0xc000017b
STATUS_KEY_DELETED = 0xc000017c
STATUS_NO_LOG_SPACE = 0xc000017d
STATUS_TOO_MANY_SIDS = 0xc000017e
STATUS_LM_CROSS_ENCRYPTION_REQUIRED = 0xc000017f
STATUS_KEY_HAS_CHILDREN = 0xc0000180
STATUS_CHILD_MUST_BE_VOLATILE = 0xc0000181
STATUS_DEVICE_CONFIGURATION_ERROR = 0xc0000182
STATUS_DRIVER_INTERNAL_ERROR = 0xc0000183
STATUS_INVALID_DEVICE_STATE = 0xc0000184
STATUS_IO_DEVICE_ERROR = 0xc0000185
STATUS_DEVICE_PROTOCOL_ERROR = 0xc0000186
STATUS_BACKUP_CONTROLLER = 0xc0000187
STATUS_LOG_FILE_FULL = 0xc0000188
STATUS_TOO_LATE = 0xc0000189
STATUS_NO_TRUST_LSA_SECRET = 0xc000018a
STATUS_NO_TRUST_SAM_ACCOUNT = 0xc000018b
STATUS_TRUSTED_DOMAIN_FAILURE = 0xc000018c
STATUS_TRUSTED_RELATIONSHIP_FAILURE = 0xc000018d
STATUS_EVENTLOG_FILE_CORRUPT = 0xc000018e
STATUS_EVENTLOG_CANT_START = 0xc000018f
STATUS_TRUST_FAILURE = 0xc0000190
STATUS_MUTANT_LIMIT_EXCEEDED = 0xc0000191
STATUS_NETLOGON_NOT_STARTED = 0xc0000192
STATUS_ACCOUNT_EXPIRED = 0xc0000193
STATUS_POSSIBLE_DEADLOCK = 0xc0000194
STATUS_NETWORK_CREDENTIAL_CONFLICT = 0xc0000195
STATUS_REMOTE_SESSION_LIMIT = 0xc0000196
STATUS_EVENTLOG_FILE_CHANGED = 0xc0000197
STATUS_NOLOGON_INTERDOMAIN_TRUST_ACCOUNT = 0xc0000198
STATUS_NOLOGON_WORKSTATION_TRUST_ACCOUNT = 0xc0000199
STATUS_NOLOGON_SERVER_TRUST_ACCOUNT = 0xc000019a
STATUS_DOMAIN_TRUST_INCONSISTENT = 0xc000019b
STATUS_FS_DRIVER_REQUIRED = 0xc000019c
STATUS_NO_USER_SESSION_KEY = 0xc0000202
STATUS_USER_SESSION_DELETED = 0xc0000203
STATUS_RESOURCE_LANG_NOT_FOUND = 0xc0000204
STATUS_INSUFF_SERVER_RESOURCES = 0xc0000205
STATUS_INVALID_BUFFER_SIZE = 0xc0000206
STATUS_INVALID_ADDRESS_COMPONENT = 0xc0000207
STATUS_INVALID_ADDRESS_WILDCARD = 0xc0000208
STATUS_TOO_MANY_ADDRESSES = 0xc0000209
STATUS_ADDRESS_ALREADY_EXISTS = 0xc000020a
STATUS_ADDRESS_CLOSED = 0xc000020b
STATUS_CONNECTION_DISCONNECTED = 0xc000020c
STATUS_CONNECTION_RESET = 0xc000020d
STATUS_TOO_MANY_NODES = 0xc000020e
STATUS_TRANSACTION_ABORTED = 0xc000020f
STATUS_TRANSACTION_TIMED_OUT = 0xc0000210
STATUS_TRANSACTION_NO_RELEASE = 0xc0000211
STATUS_TRANSACTION_NO_MATCH = 0xc0000212
STATUS_TRANSACTION_RESPONDED = 0xc0000213
STATUS_TRANSACTION_INVALID_ID = 0xc0000214
STATUS_TRANSACTION_INVALID_TYPE = 0xc0000215
STATUS_NOT_SERVER_SESSION = 0xc0000216
STATUS_NOT_CLIENT_SESSION = 0xc0000217
STATUS_CANNOT_LOAD_REGISTRY_FILE = 0xc0000218
STATUS_DEBUG_ATTACH_FAILED = 0xc0000219
STATUS_SYSTEM_PROCESS_TERMINATED = 0xc000021a
STATUS_DATA_NOT_ACCEPTED = 0xc000021b
STATUS_NO_BROWSER_SERVERS_FOUND = 0xc000021c
STATUS_VDM_HARD_ERROR = 0xc000021d
STATUS_DRIVER_CANCEL_TIMEOUT = 0xc000021e
STATUS_REPLY_MESSAGE_MISMATCH = 0xc000021f
STATUS_MAPPED_ALIGNMENT = 0xc0000220
STATUS_IMAGE_CHECKSUM_MISMATCH = 0xc0000221
STATUS_LOST_WRITEBEHIND_DATA = 0xc0000222
STATUS_CLIENT_SERVER_PARAMETERS_INVALID = 0xc0000223
STATUS_PASSWORD_MUST_CHANGE = 0xc0000224
STATUS_NOT_FOUND = 0xc0000225
STATUS_NOT_TINY_STREAM = 0xc0000226
STATUS_RECOVERY_FAILURE = 0xc0000227
STATUS_STACK_OVERFLOW_READ = 0xc0000228
STATUS_FAIL_CHECK = 0xc0000229
STATUS_DUPLICATE_OBJECTID = 0xc000022a
STATUS_OBJECTID_EXISTS = 0xc000022b
STATUS_CONVERT_TO_LARGE = 0xc000022c
STATUS_RETRY = 0xc000022d
STATUS_FOUND_OUT_OF_SCOPE = 0xc000022e
STATUS_ALLOCATE_BUCKET = 0xc000022f
STATUS_PROPSET_NOT_FOUND = 0xc0000230
STATUS_MARSHALL_OVERFLOW = 0xc0000231
STATUS_INVALID_VARIANT = 0xc0000232
STATUS_DOMAIN_CONTROLLER_NOT_FOUND = 0xc0000233
STATUS_ACCOUNT_LOCKED_OUT = 0xc0000234
STATUS_HANDLE_NOT_CLOSABLE = 0xc0000235
STATUS_CONNECTION_REFUSED = 0xc0000236
STATUS_GRACEFUL_DISCONNECT = 0xc0000237
STATUS_ADDRESS_ALREADY_ASSOCIATED = 0xc0000238
STATUS_ADDRESS_NOT_ASSOCIATED = 0xc0000239
STATUS_CONNECTION_INVALID = 0xc000023a
STATUS_CONNECTION_ACTIVE = 0xc000023b
STATUS_NETWORK_UNREACHABLE = 0xc000023c
STATUS_HOST_UNREACHABLE = 0xc000023d
STATUS_PROTOCOL_UNREACHABLE = 0xc000023e
STATUS_PORT_UNREACHABLE = 0xc000023f
STATUS_REQUEST_ABORTED = 0xc0000240
STATUS_CONNECTION_ABORTED = 0xc0000241
STATUS_BAD_COMPRESSION_BUFFER = 0xc0000242
STATUS_USER_MAPPED_FILE = 0xc0000243
STATUS_AUDIT_FAILED = 0xc0000244
STATUS_TIMER_RESOLUTION_NOT_SET = 0xc0000245
STATUS_CONNECTION_COUNT_LIMIT = 0xc0000246
STATUS_LOGIN_TIME_RESTRICTION = 0xc0000247
STATUS_LOGIN_WKSTA_RESTRICTION = 0xc0000248
STATUS_IMAGE_MP_UP_MISMATCH = 0xc0000249
STATUS_INSUFFICIENT_LOGON_INFO = 0xc0000250
STATUS_BAD_DLL_ENTRYPOINT = 0xc0000251
STATUS_BAD_SERVICE_ENTRYPOINT = 0xc0000252
STATUS_LPC_REPLY_LOST = 0xc0000253
STATUS_IP_ADDRESS_CONFLICT1 = 0xc0000254
STATUS_IP_ADDRESS_CONFLICT2 = 0xc0000255
STATUS_REGISTRY_QUOTA_LIMIT = 0xc0000256
STATUS_PATH_NOT_COVERED = 0xc0000257
STATUS_NO_CALLBACK_ACTIVE = 0xc0000258
STATUS_LICENSE_QUOTA_EXCEEDED = 0xc0000259
STATUS_PWD_TOO_SHORT = 0xc000025a
STATUS_PWD_TOO_RECENT = 0xc000025b
STATUS_PWD_HISTORY_CONFLICT = 0xc000025c
STATUS_PLUGPLAY_NO_DEVICE = 0xc000025e
STATUS_UNSUPPORTED_COMPRESSION = 0xc000025f
STATUS_INVALID_HW_PROFILE = 0xc0000260
STATUS_INVALID_PLUGPLAY_DEVICE_PATH = 0xc0000261
STATUS_DRIVER_ORDINAL_NOT_FOUND = 0xc0000262
STATUS_DRIVER_ENTRYPOINT_NOT_FOUND = 0xc0000263
STATUS_RESOURCE_NOT_OWNED = 0xc0000264
STATUS_TOO_MANY_LINKS = 0xc0000265
STATUS_QUOTA_LIST_INCONSISTENT = 0xc0000266
STATUS_FILE_IS_OFFLINE = 0xc0000267
STATUS_EVALUATION_EXPIRATION = 0xc0000268
STATUS_ILLEGAL_DLL_RELOCATION = 0xc0000269
STATUS_LICENSE_VIOLATION = 0xc000026a
STATUS_DLL_INIT_FAILED_LOGOFF = 0xc000026b
STATUS_DRIVER_UNABLE_TO_LOAD = 0xc000026c
STATUS_DFS_UNAVAILABLE = 0xc000026d
STATUS_VOLUME_DISMOUNTED = 0xc000026e
STATUS_WX86_INTERNAL_ERROR = 0xc000026f
STATUS_WX86_FLOAT_STACK_CHECK = 0xc0000270
STATUS_VALIDATE_CONTINUE = 0xc0000271
STATUS_NO_MATCH = 0xc0000272
STATUS_NO_MORE_MATCHES = 0xc0000273
STATUS_NOT_A_REPARSE_POINT = 0xc0000275
STATUS_IO_REPARSE_TAG_INVALID = 0xc0000276
STATUS_IO_REPARSE_TAG_MISMATCH = 0xc0000277
STATUS_IO_REPARSE_DATA_INVALID = 0xc0000278
STATUS_IO_REPARSE_TAG_NOT_HANDLED = 0xc0000279
STATUS_REPARSE_POINT_NOT_RESOLVED = 0xc0000280
STATUS_DIRECTORY_IS_A_REPARSE_POINT = 0xc0000281
STATUS_RANGE_LIST_CONFLICT = 0xc0000282
STATUS_SOURCE_ELEMENT_EMPTY = 0xc0000283
STATUS_DESTINATION_ELEMENT_FULL = 0xc0000284
STATUS_ILLEGAL_ELEMENT_ADDRESS = 0xc0000285
STATUS_MAGAZINE_NOT_PRESENT = 0xc0000286
STATUS_REINITIALIZATION_NEEDED = 0xc0000287
STATUS_ENCRYPTION_FAILED = 0xc000028a
STATUS_DECRYPTION_FAILED = 0xc000028b
STATUS_RANGE_NOT_FOUND = 0xc000028c
STATUS_NO_RECOVERY_POLICY = 0xc000028d
STATUS_NO_EFS = 0xc000028e
STATUS_WRONG_EFS = 0xc000028f
STATUS_NO_USER_KEYS = 0xc0000290
STATUS_FILE_NOT_ENCRYPTED = 0xc0000291
STATUS_NOT_EXPORT_FORMAT = 0xc0000292
STATUS_FILE_ENCRYPTED = 0xc0000293
STATUS_WMI_GUID_NOT_FOUND = 0xc0000295
STATUS_WMI_INSTANCE_NOT_FOUND = 0xc0000296
STATUS_WMI_ITEMID_NOT_FOUND = 0xc0000297
STATUS_WMI_TRY_AGAIN = 0xc0000298
STATUS_SHARED_POLICY = 0xc0000299
STATUS_POLICY_OBJECT_NOT_FOUND = 0xc000029a
STATUS_POLICY_ONLY_IN_DS = 0xc000029b
STATUS_VOLUME_NOT_UPGRADED = 0xc000029c
STATUS_REMOTE_STORAGE_NOT_ACTIVE = 0xc000029d
STATUS_REMOTE_STORAGE_MEDIA_ERROR = 0xc000029e
STATUS_NO_TRACKING_SERVICE = 0xc000029f
STATUS_SERVER_SID_MISMATCH = 0xc00002a0
STATUS_DS_NO_ATTRIBUTE_OR_VALUE = 0xc00002a1
STATUS_DS_INVALID_ATTRIBUTE_SYNTAX = 0xc00002a2
STATUS_DS_ATTRIBUTE_TYPE_UNDEFINED = 0xc00002a3
STATUS_DS_ATTRIBUTE_OR_VALUE_EXISTS = 0xc00002a4
STATUS_DS_BUSY = 0xc00002a5
STATUS_DS_UNAVAILABLE = 0xc00002a6
STATUS_DS_NO_RIDS_ALLOCATED = 0xc00002a7
STATUS_DS_NO_MORE_RIDS = 0xc00002a8
STATUS_DS_INCORRECT_ROLE_OWNER = 0xc00002a9
STATUS_DS_RIDMGR_INIT_ERROR = 0xc00002aa
STATUS_DS_OBJ_CLASS_VIOLATION = 0xc00002ab
STATUS_DS_CANT_ON_NON_LEAF = 0xc00002ac
STATUS_DS_CANT_ON_RDN = 0xc00002ad
STATUS_DS_CANT_MOD_OBJ_CLASS = 0xc00002ae
STATUS_DS_CROSS_DOM_MOVE_FAILED = 0xc00002af
STATUS_DS_GC_NOT_AVAILABLE = 0xc00002b0
STATUS_DIRECTORY_SERVICE_REQUIRED = 0xc00002b1
STATUS_REPARSE_ATTRIBUTE_CONFLICT = 0xc00002b2
STATUS_CANT_ENABLE_DENY_ONLY = 0xc00002b3
STATUS_FLOAT_MULTIPLE_FAULTS = 0xc00002b4
STATUS_FLOAT_MULTIPLE_TRAPS = 0xc00002b5
STATUS_DEVICE_REMOVED = 0xc00002b6
STATUS_JOURNAL_DELETE_IN_PROGRESS = 0xc00002b7
STATUS_JOURNAL_NOT_ACTIVE = 0xc00002b8
STATUS_NOINTERFACE = 0xc00002b9
STATUS_DS_ADMIN_LIMIT_EXCEEDED = 0xc00002c1
STATUS_DRIVER_FAILED_SLEEP = 0xc00002c2
STATUS_MUTUAL_AUTHENTICATION_FAILED = 0xc00002c3
STATUS_CORRUPT_SYSTEM_FILE = 0xc00002c4
STATUS_DATATYPE_MISALIGNMENT_ERROR = 0xc00002c5
STATUS_WMI_READ_ONLY = 0xc00002c6
STATUS_WMI_SET_FAILURE = 0xc00002c7
STATUS_COMMITMENT_MINIMUM = 0xc00002c8
STATUS_REG_NAT_CONSUMPTION = 0xc00002c9
STATUS_TRANSPORT_FULL = 0xc00002ca
STATUS_DS_SAM_INIT_FAILURE = 0xc00002cb
STATUS_ONLY_IF_CONNECTED = 0xc00002cc
STATUS_DS_SENSITIVE_GROUP_VIOLATION = 0xc00002cd
STATUS_PNP_RESTART_ENUMERATION = 0xc00002ce
STATUS_JOURNAL_ENTRY_DELETED = 0xc00002cf
STATUS_DS_CANT_MOD_PRIMARYGROUPID = 0xc00002d0
STATUS_SYSTEM_IMAGE_BAD_SIGNATURE = 0xc00002d1
STATUS_PNP_REBOOT_REQUIRED = 0xc00002d2
STATUS_POWER_STATE_INVALID = 0xc00002d3
STATUS_DS_INVALID_GROUP_TYPE = 0xc00002d4
STATUS_DS_NO_NEST_GLOBALGROUP_IN_MIXEDDOMAIN = 0xc00002d5
STATUS_DS_NO_NEST_LOCALGROUP_IN_MIXEDDOMAIN = 0xc00002d6
STATUS_DS_GLOBAL_CANT_HAVE_LOCAL_MEMBER = 0xc00002d7
STATUS_DS_GLOBAL_CANT_HAVE_UNIVERSAL_MEMBER = 0xc00002d8
STATUS_DS_UNIVERSAL_CANT_HAVE_LOCAL_MEMBER = 0xc00002d9
STATUS_DS_GLOBAL_CANT_HAVE_CROSSDOMAIN_MEMBER = 0xc00002da
STATUS_DS_LOCAL_CANT_HAVE_CROSSDOMAIN_LOCAL_MEMBER = 0xc00002db
STATUS_DS_HAVE_PRIMARY_MEMBERS = 0xc00002dc
STATUS_WMI_NOT_SUPPORTED = 0xc00002dd
STATUS_INSUFFICIENT_POWER = 0xc00002de
STATUS_SAM_NEED_BOOTKEY_PASSWORD = 0xc00002df
STATUS_SAM_NEED_BOOTKEY_FLOPPY = 0xc00002e0
STATUS_DS_CANT_START = 0xc00002e1
STATUS_DS_INIT_FAILURE = 0xc00002e2
STATUS_SAM_INIT_FAILURE = 0xc00002e3
STATUS_DS_GC_REQUIRED = 0xc00002e4
STATUS_DS_LOCAL_MEMBER_OF_LOCAL_ONLY = 0xc00002e5
STATUS_DS_NO_FPO_IN_UNIVERSAL_GROUPS = 0xc00002e6
STATUS_DS_MACHINE_ACCOUNT_QUOTA_EXCEEDED = 0xc00002e7
STATUS_MULTIPLE_FAULT_VIOLATION = 0xc00002e8
STATUS_CURRENT_DOMAIN_NOT_ALLOWED = 0xc00002e9
STATUS_CANNOT_MAKE = 0xc00002ea
STATUS_SYSTEM_SHUTDOWN = 0xc00002eb
STATUS_DS_INIT_FAILURE_CONSOLE = 0xc00002ec
STATUS_DS_SAM_INIT_FAILURE_CONSOLE = 0xc00002ed
STATUS_UNFINISHED_CONTEXT_DELETED = 0xc00002ee
STATUS_NO_TGT_REPLY = 0xc00002ef
STATUS_OBJECTID_NOT_FOUND = 0xc00002f0
STATUS_NO_IP_ADDRESSES = 0xc00002f1
STATUS_WRONG_CREDENTIAL_HANDLE = 0xc00002f2
STATUS_CRYPTO_SYSTEM_INVALID = 0xc00002f3
STATUS_MAX_REFERRALS_EXCEEDED = 0xc00002f4
STATUS_MUST_BE_KDC = 0xc00002f5
STATUS_STRONG_CRYPTO_NOT_SUPPORTED = 0xc00002f6
STATUS_TOO_MANY_PRINCIPALS = 0xc00002f7
STATUS_NO_PA_DATA = 0xc00002f8
STATUS_PKINIT_NAME_MISMATCH = 0xc00002f9
STATUS_SMARTCARD_LOGON_REQUIRED = 0xc00002fa
STATUS_KDC_INVALID_REQUEST = 0xc00002fb
STATUS_KDC_UNABLE_TO_REFER = 0xc00002fc
STATUS_KDC_UNKNOWN_ETYPE = 0xc00002fd
STATUS_SHUTDOWN_IN_PROGRESS = 0xc00002fe
STATUS_SERVER_SHUTDOWN_IN_PROGRESS = 0xc00002ff
STATUS_NOT_SUPPORTED_ON_SBS = 0xc0000300
STATUS_WMI_GUID_DISCONNECTED = 0xc0000301
STATUS_WMI_ALREADY_DISABLED = 0xc0000302
STATUS_WMI_ALREADY_ENABLED = 0xc0000303
STATUS_MFT_TOO_FRAGMENTED = 0xc0000304
STATUS_COPY_PROTECTION_FAILURE = 0xc0000305
STATUS_CSS_AUTHENTICATION_FAILURE = 0xc0000306
STATUS_CSS_KEY_NOT_PRESENT = 0xc0000307
STATUS_CSS_KEY_NOT_ESTABLISHED = 0xc0000308
STATUS_CSS_SCRAMBLED_SECTOR = 0xc0000309
STATUS_CSS_REGION_MISMATCH = 0xc000030a
STATUS_CSS_RESETS_EXHAUSTED = 0xc000030b
STATUS_PKINIT_FAILURE = 0xc0000320
STATUS_SMARTCARD_SUBSYSTEM_FAILURE = 0xc0000321
STATUS_NO_KERB_KEY = 0xc0000322
STATUS_HOST_DOWN = 0xc0000350
STATUS_UNSUPPORTED_PREAUTH = 0xc0000351
STATUS_EFS_ALG_BLOB_TOO_BIG = 0xc0000352
STATUS_PORT_NOT_SET = 0xc0000353
STATUS_DEBUGGER_INACTIVE = 0xc0000354
STATUS_DS_VERSION_CHECK_FAILURE = 0xc0000355
STATUS_AUDITING_DISABLED = 0xc0000356
STATUS_PRENT4_MACHINE_ACCOUNT = 0xc0000357
STATUS_DS_AG_CANT_HAVE_UNIVERSAL_MEMBER = 0xc0000358
STATUS_INVALID_IMAGE_WIN_32 = 0xc0000359
STATUS_INVALID_IMAGE_WIN_64 = 0xc000035a
STATUS_BAD_BINDINGS = 0xc000035b
STATUS_NETWORK_SESSION_EXPIRED = 0xc000035c
STATUS_APPHELP_BLOCK = 0xc000035d
STATUS_ALL_SIDS_FILTERED = 0xc000035e
STATUS_NOT_SAFE_MODE_DRIVER = 0xc000035f
STATUS_ACCESS_DISABLED_BY_POLICY_DEFAULT = 0xc0000361
STATUS_ACCESS_DISABLED_BY_POLICY_PATH = 0xc0000362
STATUS_ACCESS_DISABLED_BY_POLICY_PUBLISHER = 0xc0000363
STATUS_ACCESS_DISABLED_BY_POLICY_OTHER = 0xc0000364
STATUS_FAILED_DRIVER_ENTRY = 0xc0000365
STATUS_DEVICE_ENUMERATION_ERROR = 0xc0000366
STATUS_WAIT_FOR_OPLOCK = 0x00000367
STATUS_MOUNT_POINT_NOT_RESOLVED = 0xc0000368
STATUS_INVALID_DEVICE_OBJECT_PARAMETER = 0xc0000369
STATUS_MCA_OCCURED = 0xc000036a
STATUS_DRIVER_BLOCKED_CRITICAL = 0xc000036b
STATUS_DRIVER_BLOCKED = 0xc000036c
STATUS_DRIVER_DATABASE_ERROR = 0xc000036d
STATUS_SYSTEM_HIVE_TOO_LARGE = 0xc000036e
STATUS_INVALID_IMPORT_OF_NON_DLL = 0xc000036f
STATUS_SMARTCARD_WRONG_PIN = 0xc0000380
STATUS_SMARTCARD_CARD_BLOCKED = 0xc0000381
STATUS_SMARTCARD_CARD_NOT_AUTHENTICATED = 0xc0000382
STATUS_SMARTCARD_NO_CARD = 0xc0000383
STATUS_SMARTCARD_NO_KEY_CONTAINER = 0xc0000384
STATUS_SMARTCARD_NO_CERTIFICATE = 0xc0000385
STATUS_SMARTCARD_NO_KEYSET = 0xc0000386
STATUS_SMARTCARD_IO_ERROR = 0xc0000387
STATUS_DOWNGRADE_DETECTED = 0xc0000388
STATUS_SMARTCARD_CERT_REVOKED = 0xc0000389
STATUS_ISSUING_CA_UNTRUSTED = 0xc000038a
STATUS_REVOCATION_OFFLINE_C = 0xc000038b
STATUS_PKINIT_CLIENT_FAILURE = 0xc000038c
STATUS_SMARTCARD_CERT_EXPIRED = 0xc000038d
STATUS_DRIVER_FAILED_PRIOR_UNLOAD = 0xc000038e
STATUS_SMARTCARD_SILENT_CONTEXT = 0xc000038f
STATUS_PER_USER_TRUST_QUOTA_EXCEEDED = 0xc0000401
STATUS_ALL_USER_TRUST_QUOTA_EXCEEDED = 0xc0000402
STATUS_USER_DELETE_TRUST_QUOTA_EXCEEDED = 0xc0000403
STATUS_DS_NAME_NOT_UNIQUE = 0xc0000404
STATUS_DS_DUPLICATE_ID_FOUND = 0xc0000405
STATUS_DS_GROUP_CONVERSION_ERROR = 0xc0000406
STATUS_VOLSNAP_PREPARE_HIBERNATE = 0xc0000407
STATUS_USER2USER_REQUIRED = 0xc0000408
STATUS_STACK_BUFFER_OVERRUN = 0xc0000409
STATUS_NO_S4U_PROT_SUPPORT = 0xc000040a
STATUS_CROSSREALM_DELEGATION_FAILURE = 0xc000040b
STATUS_REVOCATION_OFFLINE_KDC = 0xc000040c
STATUS_ISSUING_CA_UNTRUSTED_KDC = 0xc000040d
STATUS_KDC_CERT_EXPIRED = 0xc000040e
STATUS_KDC_CERT_REVOKED = 0xc000040f
STATUS_PARAMETER_QUOTA_EXCEEDED = 0xc0000410
STATUS_HIBERNATION_FAILURE = 0xc0000411
STATUS_DELAY_LOAD_FAILED = 0xc0000412
STATUS_AUTHENTICATION_FIREWALL_FAILED = 0xc0000413
STATUS_VDM_DISALLOWED = 0xc0000414
STATUS_HUNG_DISPLAY_DRIVER_THREAD = 0xc0000415
STATUS_INSUFFICIENT_RESOURCE_FOR_SPECIFIED_SHARED_SECTION_SIZE = 0xc0000416
STATUS_INVALID_CRUNTIME_PARAMETER = 0xc0000417
STATUS_NTLM_BLOCKED = 0xc0000418
STATUS_ASSERTION_FAILURE = 0xc0000420
STATUS_VERIFIER_STOP = 0xc0000421
STATUS_CALLBACK_POP_STACK = 0xc0000423
STATUS_INCOMPATIBLE_DRIVER_BLOCKED = 0xc0000424
STATUS_HIVE_UNLOADED = 0xc0000425
STATUS_COMPRESSION_DISABLED = 0xc0000426
STATUS_FILE_SYSTEM_LIMITATION = 0xc0000427
STATUS_INVALID_IMAGE_HASH = 0xc0000428
STATUS_NOT_CAPABLE = 0xc0000429
STATUS_REQUEST_OUT_OF_SEQUENCE = 0xc000042a
STATUS_IMPLEMENTATION_LIMIT = 0xc000042b
STATUS_ELEVATION_REQUIRED = 0xc000042c
STATUS_BEYOND_VDL = 0xc0000432
STATUS_ENCOUNTERED_WRITE_IN_PROGRESS = 0xc0000433
STATUS_PTE_CHANGED = 0xc0000434
STATUS_PURGE_FAILED = 0xc0000435
STATUS_CRED_REQUIRES_CONFIRMATION = 0xc0000440
STATUS_CS_ENCRYPTION_INVALID_SERVER_RESPONSE = 0xc0000441
STATUS_CS_ENCRYPTION_UNSUPPORTED_SERVER = 0xc0000442
STATUS_CS_ENCRYPTION_EXISTING_ENCRYPTED_FILE = 0xc0000443
STATUS_CS_ENCRYPTION_NEW_ENCRYPTED_FILE = 0xc0000444
STATUS_CS_ENCRYPTION_FILE_NOT_CSE = 0xc0000445
STATUS_INVALID_LABEL = 0xc0000446
STATUS_DRIVER_PROCESS_TERMINATED = 0xc0000450
STATUS_AMBIGUOUS_SYSTEM_DEVICE = 0xc0000451
STATUS_SYSTEM_DEVICE_NOT_FOUND = 0xc0000452
STATUS_RESTART_BOOT_APPLICATION = 0xc0000453
STATUS_INVALID_TASK_NAME = 0xc0000500
STATUS_INVALID_TASK_INDEX = 0xc0000501
STATUS_THREAD_ALREADY_IN_TASK = 0xc0000502
STATUS_CALLBACK_BYPASS = 0xc0000503
STATUS_PORT_CLOSED = 0xc0000700
STATUS_MESSAGE_LOST = 0xc0000701
STATUS_INVALID_MESSAGE = 0xc0000702
STATUS_REQUEST_CANCELED = 0xc0000703
STATUS_RECURSIVE_DISPATCH = 0xc0000704
STATUS_LPC_RECEIVE_BUFFER_EXPECTED = 0xc0000705
STATUS_LPC_INVALID_CONNECTION_USAGE = 0xc0000706
STATUS_LPC_REQUESTS_NOT_ALLOWED = 0xc0000707
STATUS_RESOURCE_IN_USE = 0xc0000708
STATUS_HARDWARE_MEMORY_ERROR = 0xc0000709
STATUS_THREADPOOL_HANDLE_EXCEPTION = 0xc000070a
STATUS_THREADPOOL_SET_EVENT_ON_COMPLETION_FAILED = 0xc000070b
STATUS_THREADPOOL_RELEASE_SEMAPHORE_ON_COMPLETION_FAILED = 0xc000070c
STATUS_THREADPOOL_RELEASE_MUTEX_ON_COMPLETION_FAILED = 0xc000070d
STATUS_THREADPOOL_FREE_LIBRARY_ON_COMPLETION_FAILED = 0xc000070e
STATUS_THREADPOOL_RELEASED_DURING_OPERATION = 0xc000070f
STATUS_CALLBACK_RETURNED_WHILE_IMPERSONATING = 0xc0000710
STATUS_APC_RETURNED_WHILE_IMPERSONATING = 0xc0000711
STATUS_PROCESS_IS_PROTECTED = 0xc0000712
STATUS_MCA_EXCEPTION = 0xc0000713
STATUS_CERTIFICATE_MAPPING_NOT_UNIQUE = 0xc0000714
STATUS_SYMLINK_CLASS_DISABLED = 0xc0000715
STATUS_INVALID_IDN_NORMALIZATION = 0xc0000716
STATUS_NO_UNICODE_TRANSLATION = 0xc0000717
STATUS_ALREADY_REGISTERED = 0xc0000718
STATUS_CONTEXT_MISMATCH = 0xc0000719
STATUS_PORT_ALREADY_HAS_COMPLETION_LIST = 0xc000071a
STATUS_CALLBACK_RETURNED_THREAD_PRIORITY = 0xc000071b
STATUS_INVALID_THREAD = 0xc000071c
STATUS_CALLBACK_RETURNED_TRANSACTION = 0xc000071d
STATUS_CALLBACK_RETURNED_LDR_LOCK = 0xc000071e
STATUS_CALLBACK_RETURNED_LANG = 0xc000071f
STATUS_CALLBACK_RETURNED_PRI_BACK = 0xc0000720
STATUS_CALLBACK_RETURNED_THREAD_AFFINITY = 0xc0000721
STATUS_DISK_REPAIR_DISABLED = 0xc0000800
STATUS_DS_DOMAIN_RENAME_IN_PROGRESS = 0xc0000801
STATUS_DISK_QUOTA_EXCEEDED = 0xc0000802
STATUS_CONTENT_BLOCKED = 0xc0000804
STATUS_BAD_CLUSTERS = 0xc0000805
STATUS_VOLUME_DIRTY = 0xc0000806
STATUS_FILE_CHECKED_OUT = 0xc0000901
STATUS_CHECKOUT_REQUIRED = 0xc0000902
STATUS_BAD_FILE_TYPE = 0xc0000903
STATUS_FILE_TOO_LARGE = 0xc0000904
STATUS_FORMS_AUTH_REQUIRED = 0xc0000905
STATUS_VIRUS_INFECTED = 0xc0000906
STATUS_VIRUS_DELETED = 0xc0000907
STATUS_BAD_MCFG_TABLE = 0xc0000908
STATUS_WOW_ASSERTION = 0xc0009898
RPC_NT_INVALID_STRING_BINDING = 0xc0020001
RPC_NT_WRONG_KIND_OF_BINDING = 0xc0020002
RPC_NT_INVALID_BINDING = 0xc0020003
RPC_NT_PROTSEQ_NOT_SUPPORTED = 0xc0020004
RPC_NT_INVALID_RPC_PROTSEQ = 0xc0020005
RPC_NT_INVALID_STRING_UUID = 0xc0020006
RPC_NT_INVALID_ENDPOINT_FORMAT = 0xc0020007
RPC_NT_INVALID_NET_ADDR = 0xc0020008
RPC_NT_NO_ENDPOINT_FOUND = 0xc0020009
RPC_NT_INVALID_TIMEOUT = 0xc002000a
RPC_NT_OBJECT_NOT_FOUND = 0xc002000b
RPC_NT_ALREADY_REGISTERED = 0xc002000c
RPC_NT_TYPE_ALREADY_REGISTERED = 0xc002000d
RPC_NT_ALREADY_LISTENING = 0xc002000e
RPC_NT_NO_PROTSEQS_REGISTERED = 0xc002000f
RPC_NT_NOT_LISTENING = 0xc0020010
RPC_NT_UNKNOWN_MGR_TYPE = 0xc0020011
RPC_NT_UNKNOWN_IF = 0xc0020012
RPC_NT_NO_BINDINGS = 0xc0020013
RPC_NT_NO_PROTSEQS = 0xc0020014
RPC_NT_CANT_CREATE_ENDPOINT = 0xc0020015
RPC_NT_OUT_OF_RESOURCES = 0xc0020016
RPC_NT_SERVER_UNAVAILABLE = 0xc0020017
RPC_NT_SERVER_TOO_BUSY = 0xc0020018
RPC_NT_INVALID_NETWORK_OPTIONS = 0xc0020019
RPC_NT_NO_CALL_ACTIVE = 0xc002001a
RPC_NT_CALL_FAILED = 0xc002001b
RPC_NT_CALL_FAILED_DNE = 0xc002001c
RPC_NT_PROTOCOL_ERROR = 0xc002001d
RPC_NT_UNSUPPORTED_TRANS_SYN = 0xc002001f
RPC_NT_UNSUPPORTED_TYPE = 0xc0020021
RPC_NT_INVALID_TAG = 0xc0020022
RPC_NT_INVALID_BOUND = 0xc0020023
RPC_NT_NO_ENTRY_NAME = 0xc0020024
RPC_NT_INVALID_NAME_SYNTAX = 0xc0020025
RPC_NT_UNSUPPORTED_NAME_SYNTAX = 0xc0020026
RPC_NT_UUID_NO_ADDRESS = 0xc0020028
RPC_NT_DUPLICATE_ENDPOINT = 0xc0020029
RPC_NT_UNKNOWN_AUTHN_TYPE = 0xc002002a
RPC_NT_MAX_CALLS_TOO_SMALL = 0xc002002b
RPC_NT_STRING_TOO_LONG = 0xc002002c
RPC_NT_PROTSEQ_NOT_FOUND = 0xc002002d
RPC_NT_PROCNUM_OUT_OF_RANGE = 0xc002002e
RPC_NT_BINDING_HAS_NO_AUTH = 0xc002002f
RPC_NT_UNKNOWN_AUTHN_SERVICE = 0xc0020030
RPC_NT_UNKNOWN_AUTHN_LEVEL = 0xc0020031
RPC_NT_INVALID_AUTH_IDENTITY = 0xc0020032
RPC_NT_UNKNOWN_AUTHZ_SERVICE = 0xc0020033
EPT_NT_INVALID_ENTRY = 0xc0020034
EPT_NT_CANT_PERFORM_OP = 0xc0020035
EPT_NT_NOT_REGISTERED = 0xc0020036
RPC_NT_NOTHING_TO_EXPORT = 0xc0020037
RPC_NT_INCOMPLETE_NAME = 0xc0020038
RPC_NT_INVALID_VERS_OPTION = 0xc0020039
RPC_NT_NO_MORE_MEMBERS = 0xc002003a
RPC_NT_NOT_ALL_OBJS_UNEXPORTED = 0xc002003b
RPC_NT_INTERFACE_NOT_FOUND = 0xc002003c
RPC_NT_ENTRY_ALREADY_EXISTS = 0xc002003d
RPC_NT_ENTRY_NOT_FOUND = 0xc002003e
RPC_NT_NAME_SERVICE_UNAVAILABLE = 0xc002003f
RPC_NT_INVALID_NAF_ID = 0xc0020040
RPC_NT_CANNOT_SUPPORT = 0xc0020041
RPC_NT_NO_CONTEXT_AVAILABLE = 0xc0020042
RPC_NT_INTERNAL_ERROR = 0xc0020043
RPC_NT_ZERO_DIVIDE = 0xc0020044
RPC_NT_ADDRESS_ERROR = 0xc0020045
RPC_NT_FP_DIV_ZERO = 0xc0020046
RPC_NT_FP_UNDERFLOW = 0xc0020047
RPC_NT_FP_OVERFLOW = 0xc0020048
RPC_NT_CALL_IN_PROGRESS = 0xc0020049
RPC_NT_NO_MORE_BINDINGS = 0xc002004a
RPC_NT_GROUP_MEMBER_NOT_FOUND = 0xc002004b
EPT_NT_CANT_CREATE = 0xc002004c
RPC_NT_INVALID_OBJECT = 0xc002004d
RPC_NT_NO_INTERFACES = 0xc002004f
RPC_NT_CALL_CANCELLED = 0xc0020050
RPC_NT_BINDING_INCOMPLETE = 0xc0020051
RPC_NT_COMM_FAILURE = 0xc0020052
RPC_NT_UNSUPPORTED_AUTHN_LEVEL = 0xc0020053
RPC_NT_NO_PRINC_NAME = 0xc0020054
RPC_NT_NOT_RPC_ERROR = 0xc0020055
RPC_NT_SEC_PKG_ERROR = 0xc0020057
RPC_NT_NOT_CANCELLED = 0xc0020058
RPC_NT_INVALID_ASYNC_HANDLE = 0xc0020062
RPC_NT_INVALID_ASYNC_CALL = 0xc0020063
RPC_NT_NO_MORE_ENTRIES = 0xc0030001
RPC_NT_SS_CHAR_TRANS_OPEN_FAIL = 0xc0030002
RPC_NT_SS_CHAR_TRANS_SHORT_FILE = 0xc0030003
RPC_NT_SS_IN_NULL_CONTEXT = 0xc0030004
RPC_NT_SS_CONTEXT_MISMATCH = 0xc0030005
RPC_NT_SS_CONTEXT_DAMAGED = 0xc0030006
RPC_NT_SS_HANDLES_MISMATCH = 0xc0030007
RPC_NT_SS_CANNOT_GET_CALL_HANDLE = 0xc0030008
RPC_NT_NULL_REF_POINTER = 0xc0030009
RPC_NT_ENUM_VALUE_OUT_OF_RANGE = 0xc003000a
RPC_NT_BYTE_COUNT_TOO_SMALL = 0xc003000b
RPC_NT_BAD_STUB_DATA = 0xc003000c
RPC_NT_INVALID_ES_ACTION = 0xc0030059
RPC_NT_WRONG_ES_VERSION = 0xc003005a
RPC_NT_WRONG_STUB_VERSION = 0xc003005b
RPC_NT_INVALID_PIPE_OBJECT = 0xc003005c
RPC_NT_INVALID_PIPE_OPERATION = 0xc003005d
RPC_NT_WRONG_PIPE_VERSION = 0xc003005e
RPC_NT_PIPE_CLOSED = 0xc003005f
RPC_NT_PIPE_DISCIPLINE_ERROR = 0xc0030060
RPC_NT_PIPE_EMPTY = 0xc0030061
STATUS_PNP_BAD_MPS_TABLE = 0xc0040035
STATUS_PNP_TRANSLATION_FAILED = 0xc0040036
STATUS_PNP_IRQ_TRANSLATION_FAILED = 0xc0040037
STATUS_PNP_INVALID_ID = 0xc0040038
STATUS_CTX_WINSTATION_NAME_INVALID = 0xc00a0001
STATUS_CTX_INVALID_PD = 0xc00a0002
STATUS_CTX_PD_NOT_FOUND = 0xc00a0003
STATUS_CTX_CLOSE_PENDING = 0xc00a0006
STATUS_CTX_NO_OUTBUF = 0xc00a0007
STATUS_CTX_MODEM_INF_NOT_FOUND = 0xc00a0008
STATUS_CTX_INVALID_MODEMNAME = 0xc00a0009
STATUS_CTX_RESPONSE_ERROR = 0xc00a000a
STATUS_CTX_MODEM_RESPONSE_TIMEOUT = 0xc00a000b
STATUS_CTX_MODEM_RESPONSE_NO_CARRIER = 0xc00a000c
STATUS_CTX_MODEM_RESPONSE_NO_DIALTONE = 0xc00a000d
STATUS_CTX_MODEM_RESPONSE_BUSY = 0xc00a000e
STATUS_CTX_MODEM_RESPONSE_VOICE = 0xc00a000f
STATUS_CTX_TD_ERROR = 0xc00a0010
STATUS_CTX_LICENSE_CLIENT_INVALID = 0xc00a0012
STATUS_CTX_LICENSE_NOT_AVAILABLE = 0xc00a0013
STATUS_CTX_LICENSE_EXPIRED = 0xc00a0014
STATUS_CTX_WINSTATION_NOT_FOUND = 0xc00a0015
STATUS_CTX_WINSTATION_NAME_COLLISION = 0xc00a0016
STATUS_CTX_WINSTATION_BUSY = 0xc00a0017
STATUS_CTX_BAD_VIDEO_MODE = 0xc00a0018
STATUS_CTX_GRAPHICS_INVALID = 0xc00a0022
STATUS_CTX_NOT_CONSOLE = 0xc00a0024
STATUS_CTX_CLIENT_QUERY_TIMEOUT = 0xc00a0026
STATUS_CTX_CONSOLE_DISCONNECT = 0xc00a0027
STATUS_CTX_CONSOLE_CONNECT = 0xc00a0028
STATUS_CTX_SHADOW_DENIED = 0xc00a002a
STATUS_CTX_WINSTATION_ACCESS_DENIED = 0xc00a002b
STATUS_CTX_INVALID_WD = 0xc00a002e
STATUS_CTX_WD_NOT_FOUND = 0xc00a002f
STATUS_CTX_SHADOW_INVALID = 0xc00a0030
STATUS_CTX_SHADOW_DISABLED = 0xc00a0031
STATUS_RDP_PROTOCOL_ERROR = 0xc00a0032
STATUS_CTX_CLIENT_LICENSE_NOT_SET = 0xc00a0033
STATUS_CTX_CLIENT_LICENSE_IN_USE = 0xc00a0034
STATUS_CTX_SHADOW_ENDED_BY_MODE_CHANGE = 0xc00a0035
STATUS_CTX_SHADOW_NOT_RUNNING = 0xc00a0036
STATUS_CTX_LOGON_DISABLED = 0xc00a0037
STATUS_CTX_SECURITY_LAYER_ERROR = 0xc00a0038
STATUS_CLUSTER_INVALID_NODE = 0xc0130001
STATUS_CLUSTER_NODE_EXISTS = 0xc0130002
STATUS_CLUSTER_JOIN_IN_PROGRESS = 0xc0130003
STATUS_CLUSTER_NODE_NOT_FOUND = 0xc0130004
STATUS_CLUSTER_LOCAL_NODE_NOT_FOUND = 0xc0130005
STATUS_CLUSTER_NETWORK_EXISTS = 0xc0130006
STATUS_CLUSTER_NETWORK_NOT_FOUND = 0xc0130007
STATUS_CLUSTER_NETINTERFACE_EXISTS = 0xc0130008
STATUS_CLUSTER_NETINTERFACE_NOT_FOUND = 0xc0130009
STATUS_CLUSTER_INVALID_REQUEST = 0xc013000a
STATUS_CLUSTER_INVALID_NETWORK_PROVIDER = 0xc013000b
STATUS_CLUSTER_NODE_DOWN = 0xc013000c
STATUS_CLUSTER_NODE_UNREACHABLE = 0xc013000d
STATUS_CLUSTER_NODE_NOT_MEMBER = 0xc013000e
STATUS_CLUSTER_JOIN_NOT_IN_PROGRESS = 0xc013000f
STATUS_CLUSTER_INVALID_NETWORK = 0xc0130010
STATUS_CLUSTER_NO_NET_ADAPTERS = 0xc0130011
STATUS_CLUSTER_NODE_UP = 0xc0130012
STATUS_CLUSTER_NODE_PAUSED = 0xc0130013
STATUS_CLUSTER_NODE_NOT_PAUSED = 0xc0130014
STATUS_CLUSTER_NO_SECURITY_CONTEXT = 0xc0130015
STATUS_CLUSTER_NETWORK_NOT_INTERNAL = 0xc0130016
STATUS_CLUSTER_POISONED = 0xc0130017
STATUS_SXS_SECTION_NOT_FOUND = 0xc0150001
STATUS_SXS_CANT_GEN_ACTCTX = 0xc0150002
STATUS_SXS_INVALID_ACTCTXDATA_FORMAT = 0xc0150003
STATUS_SXS_ASSEMBLY_NOT_FOUND = 0xc0150004
STATUS_SXS_MANIFEST_FORMAT_ERROR = 0xc0150005
STATUS_SXS_MANIFEST_PARSE_ERROR = 0xc0150006
STATUS_SXS_ACTIVATION_CONTEXT_DISABLED = 0xc0150007
STATUS_SXS_KEY_NOT_FOUND = 0xc0150008
STATUS_SXS_VERSION_CONFLICT = 0xc0150009
STATUS_SXS_WRONG_SECTION_TYPE = 0xc015000a
STATUS_SXS_THREAD_QUERIES_DISABLED = 0xc015000b
STATUS_SXS_ASSEMBLY_MISSING = 0xc015000c
STATUS_SXS_PROCESS_DEFAULT_ALREADY_SET = 0xc015000e
STATUS_SXS_EARLY_DEACTIVATION = 0xc015000f
STATUS_SXS_INVALID_DEACTIVATION = 0xc0150010
STATUS_SXS_MULTIPLE_DEACTIVATION = 0xc0150011
STATUS_SXS_SYSTEM_DEFAULT_ACTIVATION_CONTEXT_EMPTY = 0xc0150012
STATUS_SXS_PROCESS_TERMINATION_REQUESTED = 0xc0150013
STATUS_SXS_CORRUPT_ACTIVATION_STACK = 0xc0150014
STATUS_SXS_CORRUPTION = 0xc0150015
STATUS_SXS_INVALID_IDENTITY_ATTRIBUTE_VALUE = 0xc0150016
STATUS_SXS_INVALID_IDENTITY_ATTRIBUTE_NAME = 0xc0150017
STATUS_SXS_IDENTITY_DUPLICATE_ATTRIBUTE = 0xc0150018
STATUS_SXS_IDENTITY_PARSE_ERROR = 0xc0150019
STATUS_SXS_COMPONENT_STORE_CORRUPT = 0xc015001a
STATUS_SXS_FILE_HASH_MISMATCH = 0xc015001b
STATUS_SXS_MANIFEST_IDENTITY_SAME_BUT_CONTENTS_DIFFERENT = 0xc015001c
STATUS_SXS_IDENTITIES_DIFFERENT = 0xc015001d
STATUS_SXS_ASSEMBLY_IS_NOT_A_DEPLOYMENT = 0xc015001e
STATUS_SXS_FILE_NOT_PART_OF_ASSEMBLY = 0xc015001f
STATUS_ADVANCED_INSTALLER_FAILED = 0xc0150020
STATUS_XML_ENCODING_MISMATCH = 0xc0150021
STATUS_SXS_MANIFEST_TOO_BIG = 0xc0150022
STATUS_SXS_SETTING_NOT_REGISTERED = 0xc0150023
STATUS_SXS_TRANSACTION_CLOSURE_INCOMPLETE = 0xc0150024
STATUS_SXS_PRIMITIVE_INSTALLER_FAILED = 0xc0150025
STATUS_GENERIC_COMMAND_FAILED = 0xc0150026
STATUS_SXS_FILE_HASH_MISSING = 0xc0150027
DBG_EXCEPTION_HANDLED = 0x00010001
DBG_CONTINUE = 0x00010002
DBG_TERMINATE_THREAD = 0x40010003
DBG_TERMINATE_PROCESS = 0x40010004
DBG_CONTROL_C = 0x40010005
DBG_CONTROL_BREAK = 0x40010008
DBG_COMMAND_EXCEPTION = 0x40010009
DBG_EXCEPTION_NOT_HANDLED = 0x80010001
STATUS_SUCCESS = 0x00000000



import unittest

import vivisect.tests.helpers as helpers
import vivisect.tests.utils as v_t_utils

import vdb.recon as v_rec
import vtrace
		

sym_name = {
    'windows': 'kernel32.WriteFile',
    'linux': 'libc.exit',
    'freebsd': 'libc.exit',
}

class VdbReconTests(v_t_utils.VivTest):
    '''
    Tests to verify recon hooking during execution.
    '''

    @classmethod
    def setUpClass(cls):
        cls.hello_i386_vw = helpers.getTestWorkspace('linux', 'i386',
                                                     'chgrp.llvm')
        cls.hello_amd64_vw = helpers.getTestWorkspace('linux', 'amd64',
                                                      'ls')

    def test_vdb_recon_i386(self):    
        trace = vtrace.getTrace()
        plat = trace.getMeta('Platform')
        sym = sym_name.get(plat)
		
        trace.execute(helpers.getTestPath('linux', 'i386', 'chgrp.llvm'))
        trace.setMode('RunForever', True)
        v_rec.addReconBreak(trace, sym, 'PPXPP')   
        trace.run()

        self.assertGreater(len(v_rec.getReconHits(trace)), 0)
		
    def test_vdb_recon_amd64(self):    
        trace = vtrace.getTrace()
        plat = trace.getMeta('Platform')
        sym = sym_name.get(plat)
		
        trace.execute(helpers.getTestPath('linux', 'amd64', 'ls'))
        trace.setMode('RunForever', True)
        v_rec.addReconBreak(trace, sym, 'PPXPP')
        trace.run()

        self.assertGreater(len(v_rec.getReconHits(trace)), 0)

if __name__ == '__main__':
    unittest.main()



'''
For now, all this does is rename files to their exportname and version info.
(more to come is likely)
'''

import sys
import code
import optparse
import binascii

import PE

def main():

    parser = optparse.OptionParser()
    parser.add_option('--version', dest='version', default=False, action='store_true')
    parser.add_option('--resources', dest='resources', default=False, action='store_true')

    opts, argv = parser.parse_args()

    for fname in argv:

        print('Parsing: %s' % fname)

        vsver = None
        expname = None

        pe = PE.peFromFileName(fname)

        if opts.resources:
            print('Type Nameid - rva size sample')
            for rtype, nameid, (rva, size, codepage) in pe.getResources():
                hexstr = binascii.hexlify(pe.readAtRva(rva, max(size, 8)))
                print('0x%.4x 0x%.4x - 0x%.8x 0x%.8x %s' % (rtype, nameid, rva, size, hexstr))

        if opts.version:
            vs = pe.getVS_VERSIONINFO()
            if vs is None:
                print('No VS_VERSIONINFO found!')

            else:
                keys = vs.getVersionKeys()
                keys.sort()
                for k in keys:
                    val = vs.getVersionValue(k)
                    print('%s: %r' % (k, val))

        code.interact(local=locals())


if __name__ == "__main__":
    sys.exit(main())



import logging
import binascii
import unittest

import vivisect

import vivisect.symboliks.archs.i386 as i386sym
import vivisect.symboliks.archs.amd64 as amd64sym

import vivisect.symboliks.tests.data_i386 as i386_data
import vivisect.symboliks.tests.data_amd64 as amd64_data

logger = logging.getLogger(__name__)

xlators = {
    'amd64': amd64sym.Amd64SymbolikTranslator,
    'i386': i386sym.i386SymbolikTranslator,
}

answers = {
    'amd64': amd64_data.effects,
    'i386': i386_data.effects,
}


class TranslatorTests(unittest.TestCase):

    def _run_translator_test(self, arch):
        vw = vivisect.VivWorkspace()
        vw.setMeta('Architecture', arch)
        xlate = xlators[arch](vw)
        effects = answers[arch]

        failures = []
        for name, desc in effects.items():
            opbytes, tesva, testcons, testeffs = desc
            op = vw.arch.archParseOpcode(binascii.unhexlify(opbytes), 0, 0x400)
            xlate.translateOpcode(op)

            effs = [str(e) for e in xlate.getEffects()]
            cons = [str(c) for c in xlate.getConstraints()]

            try:
                self.assertEqual(len(effs), len(testeffs))
                self.assertEqual(len(cons), len(testcons))
            except AssertionError:
                failures.append(op)
                xlate.clearEffects()
                continue

            try:
                for e in testeffs:
                    if e.startswith('SKIP'):
                        continue
                    self.assertTrue(str(e) in effs)
                for c in testcons:
                    if c.startswith('SKIP'):
                        continue
                    self.assertTrue(str(c) in cons)
            except AssertionError:
                failures.append(op)
            except Exception as e:
                logger.warning('%s failed due to exception %s', op, str(e))
                failures.append(op)
            xlate.clearEffects()

        if failures:
            msg = ', '.join([str(o) for o in failures])
            self.fail('(Arch: %s) These instructions failed to produce correct effects: %s' % (arch, msg))

    def test_amd64_instructions(self):
        self._run_translator_test('amd64')

    def test_i386_instructions(self):
        self._run_translator_test('i386')



def analyze(vw):
    done = set()
    todo = vw.getFunctions()
    while todo:
        fva = todo.pop()
        if fva in done:
            continue
        fname = vw.getName(fva)
        # XXX - get func by name
        # find the security_check_cookie function
        done.add(fva)
        if fname.startswith('security_check_cookie'):
            # parse teh first opcode in the function
            op = vw.parseOpcode(fva)
            if len(op.opers) != 2:
                return

            gscookie = None
            if op.opers[1].isDeref():
                # get the address of the cookie
                gscookie = op.opers[1].getOperAddr(op, None)

            if not gscookie:
                return

            # iterate over all references to the cookie
            for fromva, tova, rtype, rflags in vw.getXrefsTo(gscookie):
                op = vw.parseOpcode(fromva)
                cb = vw.getCodeBlock(fromva)
                if not cb:
                    continue

                va, cbsize, funcva = cb
                # if the start of the code block containing a reference to the gs cookie location isn't a
                # function then we need to make it
                # XXX - do we want to hard code mov or use a flag?? If so which flag?
                if not vw.isFunction(va) and op.mnem == 'mov':
                    vw.makeFunction(va)
                    todo.append(va)



'''
A couple useful thread related toys...
'''
import time
import logging
import threading
import functools
import collections

import envi.exc as e_exc
import envi.common as e_cmn

logger = logging.getLogger(__name__)

def firethread(func):
    '''
    A decorator which fires a thread to do the given call.

    NOTE: This means these methods may not return anything
    and callers may not expect sync behavior!
    '''
    def dothread(*args, **kwargs):
        thr = threading.Thread(target=func, args=args, kwargs=kwargs, daemon=True)
        thr.start()
        return thr
    functools.update_wrapper(dothread, func)
    return dothread


def maintthread(stime):
    '''
    A decorator which will act as a maintenance loop by calling
    back the wrapped function ( in a thread fired for this purpose )
    every "stime" seconds.
    '''
    def maintwrap(func):

        def maintloop(*args, **kwargs):
            while True:
                try:
                    func(*args, **kwargs)
                except Exception as e:
                    logger.warning('MaintThread (%s) Error: %s', args[0], e)
                time.sleep(stime)

        def dothread(*args, **kwargs):
            thr = threading.Thread(target=maintloop, args=args, kwargs=kwargs, daemon=True)
            thr.start()

        functools.update_wrapper(dothread, func)
        return dothread

    return maintwrap


class ChunkQueue:
    '''
    This is a Queue like object which returns *all* pending items
    when requested to minimize round tripping.  It's also keeps track
    of client checkins to help identify "abandonment" behaviors.
    '''
    def __init__(self, items=None, chunksize=None):
        self.shut = False
        self.last = time.time()
        self.lock = threading.Lock()
        self.event = threading.Event()
        if items is None:
            items = []
        self.items = items
        self.chunksize = chunksize

    def shutdown(self):
        self.shut = True
        self.event.set()

    def abandoned(self, dtime):
        now = time.time()
        return now > (self.last + dtime)

    def append(self, x):
        if self.shut:
            raise e_exc.QueueShutdown()
        with self.lock:
            self.items.append(x)
            self.event.set()

    def prepend(self, x):
        # NOTE: this is heavy, use judiciously
        if self.shut:
            raise e_exc.QueueShutdown()
        with self.lock:
            self.items.insert(0,x)
            self.event.set()

    def extend(self, x):
        if self.shut:
            raise e_exc.QueueShutdown()
        with self.lock:
            self.items.extend(x)
            self.event.set()

    def __len__(self):
        return len(self.items)

    def __iter__(self):
        try:
            while True:
                for i in self.get(timeout=1):
                    yield i
        except e_exc.QueueShutdown as e:
            pass

    def put(self, item):
        if self.shut:
            raise e_exc.QueueShutdown()
        self.append( item )

    def get(self, timeout=None):
        self.last = time.time()
        with self.lock:
            if self.items:
                return self._get_items()

            if self.shut:
                raise e_exc.QueueShutdown()

            # Clear the event so we can wait...
            self.event.clear()

        logger.log(e_cmn.MIRE, "self.event.wait(%d)", timeout)
        self.event.wait(timeout=timeout)
        if self.shut:
            raise e_exc.QueueShutdown()
        with self.lock:
            self.last = time.time()
            if not self.items and self.shut:
                raise e_exc.QueueShutdown()
            return self._get_items()

    def peek(self):
        return list(self.items)

    def _get_items(self):
        if self.chunksize is not None:
            ret = self.items[:self.chunksize]
            self.items = self.items[self.chunksize:]
            logger.debug("_get_items() - returning %d items (%d remaining)", len(ret), len(self.items))
        else:
            ret = self.items
            self.items = []
            logger.debug("_get_items() - returning %d items (%d remaining)", len(ret), len(self.items))
        return ret

class EnviQueue:
    '''
    A deterministic Queue that doesn't suck.
    '''
    def __init__(self, items=None):
        self.shut = False
        self.last = time.time()
        self.lock = threading.Lock()
        self.event = threading.Event()
        if items is None:
            items = []
        self.items = collections.deque(items)

    def abandoned(self, dtime):
        now = time.time()
        return now > (self.last + dtime)

    def shutdown(self):
        self.shut = True
        self.event.set()

    def append(self, x):
        if self.shut:
            raise e_exc.QueueShutdown()
        with self.lock:
            self.items.append(x)
            self.event.set()

    def prepend(self, x):
        if self.shut:
            raise e_exc.QueueShutdown()
        with self.lock:
            self.items.appendleft(x)
            self.event.set()

    def extend(self, x):
        if self.shut:
            raise e_exc.QueueShutdown()
        with self.lock:
            self.items.extend(x)
            self.event.set()

    def __len__(self):
        return len(self.items)

    def __iter__(self):
        if self.shut:
            raise e_exc.QueueShutdown()
        try:
            while True:
                ret = self.get()
                yield ret
        except e_exc.QueueShutdown as e:
            pass

    def put(self, x):
        self.append(x)

    def get(self, timeout=None):
        start = time.time()
        self.last = start

        while True:
            if self.shut:
                logger.debug("get() SHUTDOWN")
                raise e_exc.QueueShutdown()

            with self.lock:
                if self.items:
                    return self.items.popleft()

                # Clear the event so we can wait...
                self.event.clear()

            deltat = None
            if timeout:
                deltat = max(0, timeout - (time.time() - start))

            if not self.event.wait(timeout=deltat):
                return None



ord_names = {
    1:'OleUIAddVerbMenuA',
    2:'OleUICanConvertOrActivateAs',
    3:'OleUIInsertObjectA',
    4:'OleUIPasteSpecialA',
    5:'OleUIEditLinksA',
    6:'OleUIChangeIconA',
    7:'OleUIConvertA',
    8:'OleUIBusyA',
    9:'OleUIUpdateLinksA',
    10:'OleUIPromptUserA',
    11:'OleUIObjectPropertiesA',
    12:'OleUIChangeSourceA',
    13:'OleUIAddVerbMenuW',
    14:'OleUIBusyW',
    15:'OleUIChangeIconW',
    16:'OleUIChangeSourceW',
    17:'OleUIConvertW',
    18:'OleUIEditLinksW',
    19:'OleUIInsertObjectW',
    20:'OleUIObjectPropertiesW',
    21:'OleUIPasteSpecialW',
    22:'OleUIPromptUserW',
    23:'OleUIUpdateLinksW',
}



import unittest
import envi.bits as e_bits


class EnviBitsTest(unittest.TestCase):
    def test_envi_bits_masktest(self):
        self.assertTrue(e_bits.masktest('11001100')(0xcc))
        self.assertTrue(e_bits.masktest('1100110011001100')(0xcccc))
        self.assertTrue(e_bits.masktest('1100xxxx1100xxxx')(0xc2c2))
        self.assertFalse(e_bits.masktest('110011xx110011xx')(0xc2c2))

    def test_ieee754_encode(self):
        pass

    def test_byteswap(self):
        self.assertEqual(0x12345678, e_bits.byteswap(0x78563412, 4))

    def test_msb(self):
        self.assertEqual(1, e_bits.msb(0x80, 1))
        self.assertEqual(0, e_bits.msb(0x7f, 1))
        self.assertEqual(1, e_bits.msb(0x8000, 2))
        self.assertEqual(0, e_bits.msb(0x7fff, 2))
        self.assertEqual(1, e_bits.msb(0x80000000, 4))
        self.assertEqual(0, e_bits.msb(0x7fffffff, 4))
        self.assertEqual(1, e_bits.msb(0x8000000000000000, 8))
        self.assertEqual(0, e_bits.msb(0x7fffffffffffffff, 8))

        self.assertEqual(1, e_bits.msb_minus_one(0x40, 1))
        self.assertEqual(0, e_bits.msb_minus_one(0xbf, 1))
        self.assertEqual(1, e_bits.msb_minus_one(0x4000, 2))
        self.assertEqual(0, e_bits.msb_minus_one(0xbfff, 2))
        self.assertEqual(1, e_bits.msb_minus_one(0x40000000, 4))
        self.assertEqual(0, e_bits.msb_minus_one(0xbfffffff, 4))
        self.assertEqual(1, e_bits.msb_minus_one(0x4000000000000000, 8))
        self.assertEqual(0, e_bits.msb_minus_one(0xbfffffffffffffff, 8))



import time
import threading

import main

if __name__ == '__main__':
    main.waitForTest()
    thr = threading.Thread(target=time.sleep, args=(0.1,))
    thr.start()
    thr.join()
    time.sleep(0.1)
    main.exitTest()
